/*
 *	Device Registry code for the PCI bus
 *
 *	Copyright (c) 2001 Tim Jansen <tim@tjansen.de>
 *
 *  Key for physical devices is the struct pci_dev. 
 *  DeviceId is "pci/busnumber:slotnumber/vendorID:deviceID[:funcnr]".
 *  (funcnr is not set for the superdevice)
 *
 * All functions must run in user context.
 * TODO: serialize access to superdevices
 */

#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/devreg.h>
#include <linux/list.h>

#define MAX_RESOURCES 6

#define PCI_DEVREG_BUSNAME "pci"
#define PCI_DEVREG_LOCATION "PCI"

/* 
 * Struct of the superdevice for cards with more than one function
 */
struct devreg_superdevice {
	struct list_head superdevice_list;
	int bus, slot;   /* bus & slot of the superdevice */
	char deviceid[32];
	char model[80];
	char vendor[80];
	int funcnum;     /* number of functions, counter */
	int registered;  /* registered in devreg? */
	struct pci_dev *first_dev;
};

LIST_HEAD(superdevice_list);
struct semaphore superdevices_mutex;

static int write_vendorid(int *value,
			  void *context,
			  void *data)
{
	struct pci_dev *dev = (struct pci_dev*) context;
	if (!dev)
		return -ENOENT;
	*value = dev->vendor;
	return 0;
}

static int write_deviceid(int *value,
			  void *context,
			  void *data)
{
	struct pci_dev *dev = (struct pci_dev*) context;
	if (!dev)
		return -ENOENT;
	*value = dev->device;
	return 0;
}

static int write_subvendorid(int *value,
			     void *context,
			     void *data)
{
	struct pci_dev *dev = (struct pci_dev*) context;
	if (!dev)
		return -ENOENT;
	*value = dev->subsystem_vendor;
	return 0;
}

static int write_subdeviceid(int *value,
			     void *context,
			     void *data)
{
	struct pci_dev *dev = (struct pci_dev*) context;
	if (!dev)
		return -ENOENT;
	*value = dev->subsystem_device;
	return 0;
}

static int write_classid(int *value,
			 void *context,
			 void *data)
{
	u32 class_rev;
	struct pci_dev *dev = (struct pci_dev*) context;

	if (!dev)
		return -ENOENT;
	pci_read_config_dword(dev, PCI_CLASS_REVISION, &class_rev);
	*value = class_rev >> 16;
	return 0;
}

static int write_classname(char *buf,
			   void *context,
			   void *data)
{
	u32 class_rev;
	const char *text;
	struct pci_dev *dev = (struct pci_dev*) context;

	if (!dev)
		return -ENOENT;
	pci_read_config_dword(dev, PCI_CLASS_REVISION, &class_rev);
	text = pci_class_name(class_rev >> 16);
	if (!text)
		return 0;
	return sprintf(buf, "%s", text);
}

static int write_latency(int *value,
			 void *context,
			 void *data)
{
	unsigned char latency;
	struct pci_dev *dev = (struct pci_dev*) context;

	if (!dev)
		return -ENOENT;
        pci_read_config_byte (dev, PCI_LATENCY_TIMER, &latency);
	*value = latency;
	return 0;
}

static int write_mingnt(int *value,
			void *context,
			void *data)
{
	unsigned char mingnt;
	struct pci_dev *dev = (struct pci_dev*) context;

	if (!dev)
		return -ENOENT;
        pci_read_config_byte (dev, PCI_MIN_GNT, &mingnt);
	*value = mingnt;
	return 0;
}

static int write_maxlat(int *value,
			void *context,
			void *data)
{
	unsigned char maxlat;
	struct pci_dev *dev = (struct pci_dev*) context;

	if (!dev)
		return -ENOENT;
        pci_read_config_byte (dev, PCI_MAX_LAT, &maxlat);
	*value = maxlat;
	return 0;
}

static int write_irq(int *value,
		     void *context,
		     void *data)
{
	struct pci_dev *dev = (struct pci_dev*) context;
	if (!dev)
		return -ENOENT;
	*value = dev->irq;
	return 0;
}

/* if data is nonzero take io resources, otherwise mem resource */
static int res_dir_num(void *context, 
		       void *data) 
{
	struct pci_dev *dev = (struct pci_dev*) context;
	int i;
	int num = 0;

	if (!dev)
		return -ENOENT;
	for (i = 0; i < MAX_RESOURCES; i++) {
		struct resource *res = &dev->resource[i];
		if ( (res->end) && 
		     (data ? (res->flags & PCI_BASE_ADDRESS_SPACE_IO) :
		      !(res->flags & PCI_BASE_ADDRESS_SPACE_IO)) )
			num++;
	}

	return num;
}

/* if data is nonzero take io resources, otherwise mem resource */
static void *res_dir_context(int index, 
			    void *context, 
			    void *data) 
{
	struct pci_dev *dev = (struct pci_dev*) context;
	int i;
	int num = 0;

	if (!dev)
		return NULL;
	for (i = 0; i < MAX_RESOURCES; i++) {
		struct resource *res = &dev->resource[i];
		if ( (res->end) && 
		     (data ? (res->flags & PCI_BASE_ADDRESS_SPACE_IO) :
		      !(res->flags & PCI_BASE_ADDRESS_SPACE_IO)) ) {
			if (num == index)
				return res;
			num++;
		}
	}
	return NULL;
}

static int write_io_start(ulong *value,
			  void *context,
			  void *data)
{
	struct resource *res = (struct resource*) context;
	
	if (!res)
		return -ENOENT;
	*value = res->start;
	return 0;
}

static int write_io_end(ulong *value,
			void *context,
			void *data)
{
	struct resource *res = (struct resource*) context;

	if (!res)
		return -ENOENT;
	*value = res->end;
	return 0;
}

static int write_mem_base(ulong *value,
			  void *context,
			  void *data)
{
	struct resource *res = (struct resource*) context;

	if (!res)
		return -ENOENT;
	*value = res->start;
	return 0;
}

static int write_mem_end(ulong *value,
			 void *context,
			 void *data)
{
	struct resource *res = (struct resource*) context;

	if (!res)
		return -ENOENT;
	*value = res->end;
	return 0;
}

static const char *type_table[] = { 
	"unknown",
	"1m", 
	"32", 
	"64",
	NULL
};

static int write_mem_type(int *value,
			  void *context,
			  void *data)
{
	struct resource *res = (struct resource*) context;
	if (!res)
		return -ENOENT;

	switch (res->flags & PCI_BASE_ADDRESS_MEM_TYPE_MASK) { 
	case PCI_BASE_ADDRESS_MEM_TYPE_1M:
		*value = 1;
		break;
	case PCI_BASE_ADDRESS_MEM_TYPE_32:
		*value = 2; 
		break;
	case PCI_BASE_ADDRESS_MEM_TYPE_64:
		*value = 3; 
		break;		
	default:
		*value = 0;
	}
	return 0;
}

static int write_mem_prefetchable(int *value,
				  void *context,
				  void *data)
{
	struct resource *res = (struct resource*) context;
	if (!res)
		return -ENOENT;

	*value = (res->flags & PCI_BASE_ADDRESS_MEM_PREFETCH) ? 1 : 0;
	return 0;
}

static int register_pci_files (struct proc_dir_entry *dir, 
			       void *data) 
{
	struct proc_dir_entry *iodir, *memdir, *iodyndir, *memdyndir;
	#define C(x) if (!x) { \
             printk(KERN_ERR "devreg-pci: failure registering extension\n"); \
             return -ENOENT; }
	C(proc_ov_register_int(dir, "vendorid", write_vendorid));
	C(proc_ov_register_int(dir, "deviceid", write_deviceid));
	C(proc_ov_register_int(dir, "subvendorid", write_subvendorid));
	C(proc_ov_register_int(dir, "subdeviceid", write_subdeviceid));
	C(proc_ov_register_int(dir, "classid", write_classid));
	C(proc_ov_register_string(dir, "classname", write_classname));
	C(proc_ov_register_int(dir, "latency", write_latency));
	C(proc_ov_register_int(dir, "mingnt", write_mingnt));
	C(proc_ov_register_int(dir, "maxlat", write_maxlat));
	C(proc_ov_register_int(dir, "irq", write_irq));

	C((iodir = proc_ov_register_dir(dir, "io", NULL, NULL)));
	C((iodyndir = proc_ov_register_dynamic_dir(iodir, MAX_RESOURCES, 
						   res_dir_num, NULL, 
						   res_dir_context, NULL)));
	iodyndir->data = (void*) 1;
	C(proc_ov_register_ulong(iodyndir, "start", write_io_start));
	C(proc_ov_register_ulong(iodyndir, "end", write_io_end));
	
	C((memdir = proc_ov_register_dir(dir, "mem", NULL, NULL)));
	C((memdyndir = proc_ov_register_dynamic_dir(memdir, MAX_RESOURCES, 
						    res_dir_num, NULL,
						    res_dir_context, NULL)));
	C(proc_ov_register_ulong(memdyndir, "base", write_mem_base));
	C(proc_ov_register_ulong(memdyndir, "end", write_mem_end));
	C(proc_ov_register_enum(memdyndir, "type", type_table, write_mem_type));
	C(proc_ov_register_int(memdyndir, "prefetchable", write_mem_prefetchable));
	#undef C
	return 0;
}

/*
 * Returns the superdevice with in the given slot or NULL if none found
 */
static struct devreg_superdevice *get_superdevice(int bus, int slot)
{
	struct list_head *pos;	
	list_for_each(pos, &superdevice_list) {
		struct devreg_superdevice * s = 
			list_entry(pos, struct devreg_superdevice, superdevice_list);
		if ((s->bus == bus) && (s->slot == slot))
			return s;
	}
	return NULL;
}

/*
 * Creates a new superdevice and returns it, or NULL for failure
 */
static struct devreg_superdevice *create_superdevice(struct pci_dev *dev) 
{
	struct devreg_superdevice *s;

	s = (struct devreg_superdevice*) 
		kmalloc(sizeof(struct devreg_superdevice), GFP_KERNEL);
	if (!s) {
		printk(KERN_ERR 
		       "Out of memory, cannot register pci_dev in devreg");
		return NULL;
	}
	s->bus = dev->bus->number;
	s->slot = PCI_SLOT(dev->devfn);
	s->funcnum = 0;
	s->registered = 0;
	
	list_add_tail(&s->superdevice_list, &superdevice_list);
	return s;
}

/*
 * If there is a superdevice for the given struct pci_dev, it is registered.
 * If not, creates a new one.
 *
 * Takes name of the first device as model for the superdevice in devreg.
 */
static struct devreg_superdevice *register_superdevice(struct pci_dev *dev)
{
	struct pci_bus *bus = dev->bus;
	struct devreg_superdevice *s;

	s = get_superdevice(bus->number, PCI_SLOT(dev->devfn));
	if (!s) {
		s = create_superdevice(dev);
		if (!s)
			return NULL;

		sprintf(s->deviceid, 
			"pci/%d:%d/%x:%x",
			bus->number,
			PCI_SLOT(dev->devfn),
			dev->vendor,
			dev->device);

		strcpy(s->model, dev->model_name);
		strcpy(s->vendor, dev->vendor_name);
	}

	s->funcnum++;

	if (s->funcnum == 1)
		s->first_dev = dev;
	else if ((!s->registered) && (s->funcnum == 2)) {
		devreg_register_physical_device(s,
						NULL,
						NULL, 
						PCI_DEVREG_BUSNAME,
						NULL, /* type */ 
						s->deviceid,
						NULL,
						s->model,
						s->vendor, 
						PCI_DEVREG_LOCATION,
						DEVREG_LOCATION_INTERNAL);
		devreg_set_superdevice(s->first_dev, s);
		s->registered = 1;
	}

	return s;
}

/*
 * Unregisters device from superdevice, frees superdevice if neccessary.
 *
 * Takes name of the first device as model for the superdevice in devreg.
 */
static void unregister_superdevice(struct pci_dev *dev)
{
	struct devreg_superdevice *s;
	s = get_superdevice(dev->bus->number,
			    PCI_SLOT(dev->devfn));
	if (!s) 
		return;

	s->funcnum--;
	if (!s->funcnum)
		return;

	devreg_unregister_physical_device(s);
	list_del(&s->superdevice_list);
	kfree(s);
}


/*
 * Attaches a PCI device to devreg.
 * TODO: recognize AGP devices and set the bus accordingly
 */
void pci_devreg_attach_device(struct pci_dev *dev)
{
	char deviceid[32];
	struct pci_bus *bus = dev->bus;
	struct devreg_superdevice *s;

	down(&superdevices_mutex);
	s = register_superdevice(dev);
	sprintf(deviceid, 
		"pci/%d:%d/%x:%x:%x",
		bus->number,
		PCI_SLOT(dev->devfn),
		dev->vendor,
		dev->device,
		PCI_FUNC(dev->devfn));

	devreg_register_physical_device(dev,
					NULL,
					s->registered ? s : NULL,
					PCI_DEVREG_BUSNAME,
					NULL,
					deviceid,
					NULL,
					dev->model_name,
					dev->vendor_name, 
					PCI_DEVREG_LOCATION,
					DEVREG_LOCATION_INTERNAL);

	devreg_register_extension(dev, "pci", register_pci_files, NULL,
				  NULL, NULL, dev);
	up(&superdevices_mutex);
}

/*
 * Note that this function does not only unregister it. By calling
 * devreg_unregister_physical_device() it acquires a write rwlock on 
 * the devreg data, so the function wont return before all readers have
 * finished.
 */
void pci_devreg_detach_device(struct pci_dev *dev)
{
	down(&superdevices_mutex);

	devreg_unregister_physical_device(dev);
	unregister_superdevice(dev);

	up(&superdevices_mutex);
}


void __init pci_devreg_init(void)
{
	struct pci_dev *dev;

	init_MUTEX(&superdevices_mutex);

	pci_for_each_dev(dev) {
		pci_devreg_attach_device(dev);
	}
}

