/*
 *	Device Registry code for the USB
 *
 *	Copyright (c) 2001 Tim Jansen <tim@tjansen.de>
 *      (many snippets taken from drivers/usb/devices.c)     
 * 
 *  Key for physical devices is the struct usb_device. 
 *  DeviceId is 
 *  "usb/busnumber[[:1st-port-num[:2nd-port-num...]]:port/vendorID:deviceID".
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/usb.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/devreg.h>
#include "devreg_usb.h"

#define USB_DEVREG_BUSNAME "usb"
#define USB_DEVREG_LOCATION "USB"

/* limit number of <interfaces>*<alternate-settings> to 
   MAX_INTERFACES (instead of MAX_INTERFACES * MAX_ALTSETTINGS)
   to save inodes. */
#define MAXINTERFACESETTINGS USB_MAXINTERFACES 

#define USB_MAXENDPOINTS	32


static const char *_strdup(const char *str) 
{
	int l;
	char *n;
	if (!str)
		return NULL;
	l = strlen(str);
	n = kmalloc((l+1) * sizeof(char), GFP_KERNEL);
	if (!n)
		return NULL;
	strcpy(n, str);
	return n;
}

static unsigned int get_portnum(struct usb_device *dev) 
{
	int i;
	struct usb_device *pdev;
	pdev = dev->parent;
	if (!pdev)
		return 0;
	for (i = 0; i < USB_MAXCHILDREN; i++) 
		if (pdev->children[i] == dev) 
			break;
	return i;
}

static int write_u8(int *value,
		    void *context,
		    void *data)
{
	__u8 *v;
	if (!context)
		return -ENOENT;
	v = (__u8*) (context + (long) data);
	*value = *v;
	return 0;
}
static int write_u16(int *value,
		     void *context,
		     void *data)
{
	__u16 *v;
	if (!context)
		return -ENOENT;
	v = (__u16*) (context + (long) data);
	*value = *v;
	return 0;
}
static int write_bus_int(int *value,
			 void *context,
			 void *data)
{
	int *v;
	struct usb_device *dev = (struct usb_device*) context;
	if (!context)
		return -ENOENT;
	v = (int*) (((void*)dev->bus) + (long) data);
	*value = *v;
	return 0;
}
static int write_desc_u8(int *value,
			 void *context,
			 void *data)
{
	__u8 *v;
	struct usb_device *dev = (struct usb_device*) context;
	if (!context)
		return -ENOENT;
	v = (__u8*) (((void*)&dev->descriptor) + (long) data);
	*value = *v;
	return 0;
}
static int write_desc_u16(int *value,
			  void *context,
			  void *data)
{
	__u16 *v;
	struct usb_device *dev = (struct usb_device*) context;
	if (!context)
		return -ENOENT;
	v = (__u16*) (((void*)&dev->descriptor) + (long) data);
	*value = *v;
	return 0;
}


static int write_maxbwallocated(int *value,
				void *context,
				void *data)
{
	*value = FRAME_TIME_MAX_USECS_ALLOC;
	return 0;
}
static int write_port(int *value,
		      void *context,
		      void *data)
{
	struct usb_device *dev = (struct usb_device*) context;
	if (!dev)
		return -ENOENT;
	*value = get_portnum(dev);
	return 0;
}
static int write_maxports(int *value,
			  void *context,
			  void *data)
{
	struct usb_device *dev = (struct usb_device*) context;
	if (!dev)
		return -ENOENT;
	*value = dev->maxchild;
	return 0;
}
const char *speed_table[] = {
	"slow", "full", "high", NULL
};
static int write_speed(int *value,
		       void *context,
		       void *data)
{
	struct usb_device *dev = (struct usb_device*) context;
	if (!dev)
		return -ENOENT;
	*value = dev->slow ? 0 : 1;
	return 0;
}
static int write_serialnumber(char *buf,
			      void *context,
			      void *data)
{
	struct usb_device *dev = (struct usb_device*) context;
	if (!dev)
		return -ENOENT;
	if (!dev->descriptor.iSerialNumber)
		return 0;
	return usb_string(dev, 
			  dev->descriptor.iSerialNumber, 
			  buf, 
			  PROC_BLOCK_SIZE);
}

static int write_config_active(int *value,
			       void *context,
			       void *data)
{
	long a;
	struct usb_device *dev = (struct usb_device*) context;

	if (!dev)
		return -ENOENT;

	a = ((long) dev->actconfig - (long) dev->config) / 
		sizeof(struct usb_config_descriptor);
	if ((a >= 0) && (a < dev->descriptor.bNumConfigurations))
		*value = a;
	else
		*value = -1;
	return 0;
}
static int conf_dir_num(void *context, 
			void *data) 
{
	struct usb_device *dev = (struct usb_device*) context;
	if (!dev)
		return -ENOENT;
	return dev->descriptor.bNumConfigurations;
}
static void *conf_dir_context(int index, 
			      void *context, 
			      void *data) 
{
	struct usb_device *dev = (struct usb_device*) context;
	if (!dev)
		return NULL;
	if (index >= dev->descriptor.bNumConfigurations)
		return NULL;
	return &dev->config[index];	
}

static int intf_dir_num(void *context, 
			void *data) 
{
	int i;
	int n = 0;
	struct usb_config_descriptor *config = 
		(struct usb_config_descriptor*) context;
	if (!config)
		return -ENOENT;

	for (i = 0; i < config->bNumInterfaces; i++) 
		n += config->interface[i].num_altsetting;

	if (n > MAXINTERFACESETTINGS)
		return MAXINTERFACESETTINGS;
	
	return n;
}
static void *intf_dir_context(int index, 
			      void *context, 
			      void *data)
{
	int n = 0;
	int i;
	struct usb_interface *interface;
	struct usb_config_descriptor *config = 
		(struct usb_config_descriptor*) context;
	if (!config)
		return NULL;

	for (i = 0; i < config->bNumInterfaces; i++) {
		interface = &config->interface[i];
		if (index >= (n + interface->num_altsetting))
			n += interface->num_altsetting;
		else
			return &interface->altsetting[index - n];
	}

	return NULL;	
}

static int endp_dir_num(void *context, 
			void *data) 
{
	struct usb_interface_descriptor *desc =
		(struct usb_interface_descriptor*) context;
	if (!desc)
		return -ENOENT;

	return desc->bNumEndpoints;
}
static void *endp_dir_context(int index, 
			      void *context, 
			      void *data) 
{
	struct usb_interface_descriptor *desc =
		(struct usb_interface_descriptor*) context;
	if (!desc)
		return NULL;
	if (index >= desc->bNumEndpoints)
		return NULL;
	
	return &desc->endpoint[index];
}
const char *endptype_table[] = {
	"control", "isochronous", "bulk", "interrupt", NULL
};
static int write_endpoint_type(int *value,
			       void *context,
			       void *data)
{
	struct usb_endpoint_descriptor *desc = 
		(struct usb_endpoint_descriptor*) context;

	if (!desc)
		return -ENOENT;

	*value = desc->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK;
	return 0;
}
const char *endpdirection_table[] = {
	"bidirectional", "in", "out", NULL
};
static int write_endpoint_direction(int *value,
				    void *context,
				    void *data)
{
	struct usb_endpoint_descriptor *desc = 
		(struct usb_endpoint_descriptor*) context;

	if (!desc)
		return -ENOENT;

	*value = (desc->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) == USB_ENDPOINT_XFER_CONTROL
		? 0 : (desc->bEndpointAddress & USB_DIR_IN) ? 1 : 2;
	return 0;
}

static int register_usb_files (struct proc_dir_entry *dir, 
			       void *data) 
{
	struct proc_dir_entry *confdir, *confdyndir, *intfdir, *intfdyndir;
	struct proc_dir_entry *endpdir, *endpdyndir, *x;
	#define C(y) if (!(x = y)) { \
             printk(KERN_ERR "devreg-usb: failure registering extension\n"); \
             return -ENOENT; }
	#define DATA(y) x->data = y
	#define OFFS_BUS(member) (&((struct usb_bus *)0)->member)
	#define OFFS_DEVDESC(member) (&((struct usb_device_descriptor *)0)->member)
	#define OFFS_CONFDESC(member) (&((struct usb_config_descriptor *)0)->member)
	#define OFFS_INTFDESC(member) (&((struct usb_interface_descriptor *)0)->member)
	#define OFFS_ENDPDESC(member) (&((struct usb_endpoint_descriptor *)0)->member)
	#ifdef __LITTLE_ENDIAN
	int minor_offset = 0;
	int major_offset = 1;
	#else
	int minor_offset = 1;
	int major_offset = 0;
	#endif

	C(proc_ov_register_int(dir, "busnum", write_bus_int));
	DATA(OFFS_BUS(busnum));
	C(proc_ov_register_int(dir, "bwallocated", write_bus_int));
	DATA(OFFS_BUS(bandwidth_allocated));
	C(proc_ov_register_int(dir, "maxbwallocated", write_maxbwallocated));
	C(proc_ov_register_int(dir, "intrq", write_bus_int));
	DATA(OFFS_BUS(bandwidth_int_reqs));
	C(proc_ov_register_int(dir, "isorq", write_bus_int));
	DATA(OFFS_BUS(bandwidth_isoc_reqs));
	C(proc_ov_register_int(dir, "port", write_port));
	C(proc_ov_register_int(dir, "maxports", write_maxports));
	C(proc_ov_register_enum(dir, "speed", speed_table, write_speed));
	C(proc_ov_register_int(dir, "versionminor", write_desc_u8));
	DATA(OFFS_DEVDESC(bcdUSB) + minor_offset);
	C(proc_ov_register_int(dir, "versionmajor", write_desc_u8));
	DATA(OFFS_DEVDESC(bcdUSB) + major_offset);
	C(proc_ov_register_int(dir, "class", write_desc_u8));
	DATA(OFFS_DEVDESC(bDeviceClass));
	C(proc_ov_register_int(dir, "subclass", write_desc_u8));
	DATA(OFFS_DEVDESC(bDeviceSubClass));
	C(proc_ov_register_int(dir, "protocol", write_desc_u8));
	DATA(OFFS_DEVDESC(bDeviceProtocol));
	C(proc_ov_register_int(dir, "maxpacketsize", write_desc_u8));
	DATA(OFFS_DEVDESC(bMaxPacketSize0));
	C(proc_ov_register_int(dir, "vendorid", write_desc_u16));
	DATA(OFFS_DEVDESC(idVendor));
	C(proc_ov_register_int(dir, "productid", write_desc_u16));
	DATA(OFFS_DEVDESC(idProduct));
	C(proc_ov_register_int(dir, "revisionminor", write_desc_u8));
	DATA(OFFS_DEVDESC(bcdDevice) + minor_offset);
	C(proc_ov_register_int(dir, "revisionmajor", write_desc_u8));
	DATA(OFFS_DEVDESC(bcdDevice) + major_offset);
	C(proc_ov_register_string(dir, "serialnumber", write_serialnumber));


	C((confdir = proc_ov_register_dir(dir, "configurations", NULL, NULL)));
	C(proc_ov_register_int(confdir, "active", write_config_active));
	C((confdyndir = proc_ov_register_dynamic_dir(confdir, USB_MAXCONFIG, 
						     conf_dir_num, NULL, 
						     conf_dir_context, NULL)));
	C(proc_ov_register_int(confdyndir, "attributes", write_u8));
	DATA(OFFS_CONFDESC(bmAttributes));
	C(proc_ov_register_int(confdyndir, "powerhma", write_u8));
	DATA(OFFS_CONFDESC(MaxPower));

	C((intfdir = proc_ov_register_dir(confdyndir, "interfaces", NULL, NULL)));
	C((intfdyndir = proc_ov_register_dynamic_dir(intfdir, MAXINTERFACESETTINGS, 
						     intf_dir_num, NULL, 
						     intf_dir_context, NULL)));
	C(proc_ov_register_int(intfdyndir, "num", write_u8));
	DATA(OFFS_INTFDESC(bInterfaceNumber));
	C(proc_ov_register_int(intfdyndir, "class", write_u8));
	DATA(OFFS_INTFDESC(bInterfaceClass));
	C(proc_ov_register_int(intfdyndir, "subclass", write_u8));
	DATA(OFFS_INTFDESC(bInterfaceSubClass)); 
	C(proc_ov_register_int(intfdyndir, "protocol", write_u8));
	DATA(OFFS_INTFDESC(bInterfaceProtocol));
	C(proc_ov_register_int(intfdyndir, "asetting", write_u8));
	DATA(OFFS_INTFDESC(bAlternateSetting));

	C((endpdir = proc_ov_register_dir(intfdyndir, "endpoints", NULL, NULL)));
	C((endpdyndir = proc_ov_register_dynamic_dir(endpdir, USB_MAXENDPOINTS, 
						     endp_dir_num, NULL, 
						     endp_dir_context, NULL)));
	C(proc_ov_register_int(endpdyndir, "address", write_u8));
	DATA(OFFS_ENDPDESC(bEndpointAddress));
	C(proc_ov_register_int(endpdyndir, "attributes", write_u8));
	DATA(OFFS_ENDPDESC(bmAttributes));
	C(proc_ov_register_int(endpdyndir, "packetsize", write_u16));
	DATA(OFFS_ENDPDESC(wMaxPacketSize));
	C(proc_ov_register_int(endpdyndir, "interval", write_u8));
	DATA(OFFS_ENDPDESC(bInterval));
	C(proc_ov_register_enum(endpdyndir, "type", endptype_table, 
				write_endpoint_type));
	C(proc_ov_register_enum(endpdyndir, "direction", endpdirection_table, 
				 write_endpoint_direction));

	#undef C
	return 0;
}


static unsigned int write_hubpath(struct usb_device *dev, 
				  char *buf,
				  unsigned int bufsize) 
{
	unsigned int l;
	if((!dev) || (!dev->parent))
		return 0;

	l = write_hubpath(dev->parent, buf, bufsize);

	if (bufsize-l >= 10)
		l += sprintf(buf+l, ":%d", get_portnum(dev));
	return l;
}

static void write_device_version(char *version, int bcdVersion) 
{
	char *vp = version;
	int digit;

	digit = (bcdVersion >> 12) & 0xf;
	if (digit)
		*(vp++) = digit + '0';
	*(vp++) = ((bcdVersion >> 8)  & 0xf) + '0';

	*(vp++) = '.';

	*(vp++) = ((bcdVersion >> 4) & 0xf) + '0';
	digit = bcdVersion  & 0xf;
	if (digit)
		*(vp++) = digit + '0';

	*vp = 0;
}


#define BUFSIZE (PAGE_SIZE/2)
void usb_devreg_attach_device(struct usb_device *dev)
{
	char *buf;
	char *idlocation;
	const char *model;
	const char *vendor;
	const char *sernum;
	const char version[6];

	buf = (char*) __get_free_page(GFP_ATOMIC);
	if (!buf) {
		printk(KERN_ERR "Not enough memory to register usb "
		       "device in devreg!\n");
		return;
	}
	idlocation = buf + BUFSIZE;

	vendor = NULL;
	if (dev->descriptor.iManufacturer) {   
                if (usb_string(dev, dev->descriptor.iManufacturer, buf, BUFSIZE) > 0)
			vendor = _strdup(buf);
        }
	if (dev->descriptor.iProduct) {   
                if (usb_string(dev, dev->descriptor.iProduct, buf, BUFSIZE) > 0)
			model = _strdup(buf);
		else 
			model = _strdup("");
        }
	else
		model = _strdup("");
	sernum = NULL;
	if (dev->descriptor.iSerialNumber) 
                if (usb_string(dev, dev->descriptor.iSerialNumber, buf, BUFSIZE) > 0)
			sernum = _strdup(buf);

	idlocation[0] = '\0'; 
	write_hubpath(dev, idlocation, BUFSIZE-32);

	write_device_version(version, dev->descriptor.bcdDevice);
	sprintf(buf,
		"usb/%d%s/%x:%x:%s",
		dev->bus->busnum,
		idlocation,
		dev->descriptor.idVendor,
		dev->descriptor.idProduct,
		version);

	devreg_register_physical_device(dev,
					dev->parent,
					NULL, /*superdevice*/
					USB_DEVREG_BUSNAME,
					NULL,
					buf,
					sernum, 
					model, 
					vendor, 
					USB_DEVREG_LOCATION,
					DEVREG_LOCATION_EXTERNAL);
	kfree(sernum);
	free_page((unsigned long)buf);
	devreg_register_extension(dev, "usb", register_usb_files, NULL,
				  NULL, NULL, dev);

}

/*
 * Note that this function does not only unregister it. By calling
 * devreg_unregister_pysical_device() it acquires a rwlock on 
 * the devreg data, so the function wont return before all readers have
 * finished.
 */	
void usb_devreg_detach_device(struct usb_device *dev)
{
	const char *model;
	const char *vendor;

	if (!devreg_get_physical_device(dev,
					NULL,
					NULL,
					NULL,
					NULL,
					NULL,
					NULL,
					&model,
					&vendor,
					NULL,
					NULL))
		return;
	
	devreg_unregister_physical_device(dev);

	kfree(model);
	kfree(vendor);
}

static void for_all_devices(struct usb_device *root, 
		       void (*cb) (struct usb_device *)) {
	int i;

	for (i = 0; i < USB_MAXCHILDREN; i++)
		if (root->children[i])
			for_all_devices(root->children[i], cb);

	cb(root);
}

void __init usb_devreg_init(void)
{
	struct list_head *buslist;
	struct usb_bus *bus;

	for (buslist = usb_bus_list.next; buslist != &usb_bus_list; buslist = buslist->next) {
		bus = list_entry(buslist, struct usb_bus, bus_list);
		for_all_devices(bus->root_hub, usb_devreg_attach_device);
	}
}

void __exit usb_devreg_cleanup(void)
{
	struct list_head *buslist;
	struct usb_bus *bus;

	for (buslist = usb_bus_list.next; buslist != &usb_bus_list; buslist = buslist->next) {
		bus = list_entry(buslist, struct usb_bus, bus_list);
		for_all_devices(bus->root_hub, usb_devreg_detach_device);
	}
}

EXPORT_SYMBOL(usb_devreg_attach_device);
EXPORT_SYMBOL(usb_devreg_detach_device);
EXPORT_SYMBOL(usb_devreg_cleanup);
EXPORT_SYMBOL(usb_devreg_init);

