/*
 * devreg.c
 * 
 * (C) Copyright Tim Jansen 2001 <tim@tjansen.de>
 *
 * Implements the public functions from linux/devreg.h (look there for 
 * documentation) and defines helper functions and variables from 
 * devreg_t.h. 
 */

#include "proc.h"
#include "devreg_t.h"
#include <linux/kernel.h>
#include <linux/list.h>
#include <linux/init.h>
#include <linux/string.h>
#include <linux/slab.h>
#include <linux/module.h>

/* see devreg_t.h */
LIST_HEAD(devreg_devices);
volatile int devreg_change_count;
struct rw_semaphore devreg_devices_lock; 
DECLARE_WAIT_QUEUE_HEAD(devreg_change_queue);

void devreg_write_lock_devices()
{
	down_write(&devreg_devices_lock);
	devreg_change_count++;
}

void devreg_write_unlock_devices()
{
	up_write(&devreg_devices_lock);
	wake_up(&devreg_change_queue);
}

static struct devreg_phydev* _find_physical_device(void *key)
{
	struct list_head *pos;
	struct devreg_phydev *dev;

	if (!key)
		return NULL;

	list_for_each(pos, &devreg_devices) {
		dev = list_entry(pos, struct devreg_phydev, phydevs);
		if (dev->key == key)
			return dev;
	}
	 
	return NULL;
}

static const char *_strdup(const char *str) 
{
	int l;
	char *n;
	if (!str)
		return NULL;
	l = strlen(str);
	n = kmalloc((l+1) * sizeof(char), GFP_KERNEL);
	if (!n)
		return NULL;
	strcpy(n, str);
	return n;
}

void devreg_register_physical_device(void *key,
				     void *parent_key,
				     void *superdevice_key,
				     const char *bus,
				     const char *type,
				     const char *deviceid,
				     const char *deviceidsernum,
				     const char *model,
				     const char *vendor,
				     const char *location_desc,
				     enum devreg_location location)
{
	struct devreg_phydev *nd;

	if ((!key) ||
	    (!bus) || 
	    (!deviceid) || 
	    (!model) || 
	    (!location_desc)) {
		printk(KERN_ERR "devreg: devreg_register_physical_device "
		   		" argument invalid/null\n");	
	    	return;    
	}

	nd = (struct devreg_phydev*) kmalloc(sizeof(struct devreg_phydev),
					     GFP_KERNEL);
	if (!nd) {
		printk(KERN_ERR "devreg: out of memory\n");
		return;
	}
	
	nd->bus = bus;
	nd->type = _strdup(type);
	if (type && !nd->type) {
		printk(KERN_ERR "devreg: out of memory\n");
		kfree(nd);
		return;
	}
	nd->deviceidwosn = _strdup(deviceid);
	if (deviceid && !nd->deviceidwosn) {
		printk(KERN_ERR "devreg: out of memory\n");
		kfree(nd->type);
		kfree(nd);
		return;
	}
	nd->deviceidsernum = _strdup(deviceidsernum);
	if (deviceidsernum && !nd->deviceidsernum) {
		printk(KERN_ERR "devreg: out of memory\n");
		kfree(nd->deviceidwosn);
		kfree(nd->type);
		kfree(nd);
		return;
	}
	nd->model = model;
	nd->vendor = vendor;
	nd->location_desc = location_desc;
	nd->location = location;
	nd->key = key;
	nd->parent_key = parent_key;
	nd->superdevice_key = superdevice_key;

	if (!INIT_VECTOR(&nd->devfiles, devfs_handle_t, 2, GFP_KERNEL)) {
		printk(KERN_ERR "devreg: out of memory\n");
		kfree(nd->deviceidsernum);
		kfree(nd->deviceidwosn);
		kfree(nd->type);
		kfree(nd);
	}
	if (!INIT_VECTOR(&nd->extensions, struct devreg_extension, 
			2, GFP_KERNEL)) {
		printk(KERN_ERR "devreg: out of memory\n");
		FREE_VECTOR(&nd->extensions);
		kfree(nd->deviceidsernum);
		kfree(nd->deviceidwosn);
		kfree(nd->type);
		kfree(nd);
	}

	devreg_write_lock_devices();
	list_add_tail(&nd->phydevs, &devreg_devices);
	devreg_write_unlock_devices();
}


int devreg_get_physical_device(void *key,
			       void **parent_key,
			       void **superdevice_key,
			       const char **bus,
			       const char **type,
			       const char **deviceid,
			       const char **deviceidsernum,
			       const char **model,
			       const char **vendor,
			       const char **location_desc,
			       enum devreg_location *location)
{
	struct devreg_phydev* dev;

	devreg_read_lock_devices();
	dev =_find_physical_device(key);
	if (!dev) {
		devreg_read_unlock_devices();
		return 0;
	}

	if (parent_key)
		*parent_key = dev->parent_key;
	if (superdevice_key)
		*superdevice_key = dev->superdevice_key;
	if (bus)
		*bus = dev->bus;
	if (type)
		*type = dev->type;
	if (deviceid)
		*deviceid = dev->deviceidwosn;
	if (deviceidsernum)
		*deviceidsernum = dev->deviceidsernum;
	if (model)
		*model = dev->model;
	if (vendor)
		*vendor = dev->vendor;
	if (location_desc)
		*location_desc = dev->location_desc;
	if (location)
		*location = dev->location;

	devreg_read_unlock_devices();
	return 1;
}


static void _unregister_device(struct devreg_phydev *dev, int depth)
{
	struct devreg_extension_instance *ex;
	struct list_head *pos;
	struct devreg_phydev *d;
	
	if (depth <= 0) {
		printk(KERN_ERR
		       "devreg: error unregistering device, "
		       "list corrupted, maximum recursion reached\n");
		return;
	}
	
	kfree(dev->deviceidwosn);
	kfree(dev->deviceidsernum);
	kfree(dev->type);

	list_del(&dev->phydevs);

	list_for_each(pos, &devreg_devices) {
		d = list_entry(pos, struct devreg_phydev, phydevs);
		if (d->parent_key == dev->key)
			_unregister_device(d, depth-1);
	}

	list_for_each(pos, &devreg_devices) {
		d = list_entry(pos, struct devreg_phydev, phydevs);
		if (d->superdevice_key == dev->key)
			_unregister_device(d, depth-1);
	}

	VECTOR_FOR_EACH(&dev->extensions,
			ex,
			struct devreg_extension_instance) {
		if (!(--ex->extension->rc))
			devreg_free_extension(ex->extension);
	}

	FREE_VECTOR(&dev->devfiles);
	FREE_VECTOR(&dev->extensions);

	kfree(dev);
}

#define MAXIMUM_DEPTH 100

void devreg_unregister_physical_device(void *key)
{
	struct devreg_phydev *dev;

	devreg_write_lock_devices();
	
	dev = _find_physical_device(key);
	if (!dev) {
		devreg_write_unlock_devices();
		printk(KERN_WARNING
		       "devreg: tried to unregister unknown key\n");
		return;
	}

	_unregister_device(dev, MAXIMUM_DEPTH);
	devreg_write_unlock_devices();
}

static void _set_type(struct devreg_phydev *dev, const char *type)
{
	const char *t;

	t = _strdup(type);
	if ((!type) || t) {
		kfree(dev->type);
		dev->type = t;
	}
	else
		printk(KERN_WARNING "devreg: out-of-memory - "
		       "cannot change type");
}

void devreg_set_device_type(void *key, const char *type)
{
	struct devreg_phydev *dev;
	
	if (!key)
		return;

	devreg_write_lock_devices();
	dev = _find_physical_device(key);
	if (!dev) {
		devreg_write_unlock_devices();
		printk(KERN_WARNING
		       "devreg: tried to change unknown key to type %s\n", 
		       type);
		return;
	}
	_set_type(dev, type);	
	dev = _find_physical_device(dev->superdevice_key);
	if (dev && (!dev->type))
		_set_type(dev, type);
	devreg_write_unlock_devices();
}

void devreg_set_device_serial_number(void *key,
				     const char *serial_num) 
{
	struct devreg_phydev *dev;
	const char *t;

	if (!key)
		return;

	devreg_write_lock_devices();
	dev = _find_physical_device(key);
	if (!dev) {
		devreg_write_unlock_devices();
		printk(KERN_WARNING
		       "devreg: tried to change unknown key\n");
		return;
	}
	t = _strdup(serial_num);
	if (t || !serial_num) {
		kfree(dev->deviceidsernum);
		dev->deviceidsernum = t;
	}
	else
		printk(KERN_WARNING "devreg: out-of-memory - "
		       "cannot change sernum");
	devreg_write_unlock_devices();
}

void devreg_set_superdevice(void *key,
			    void *superdevice_key) 
{
	struct devreg_phydev *dev;

	if (!key)
		return;

	devreg_write_lock_devices();
	dev = _find_physical_device(key);
	if (!dev) {
		devreg_write_unlock_devices();
		printk(KERN_WARNING
		       "devreg: tried to change supderdeice of unknown key\n");
		return;
	}
	dev->superdevice_key = superdevice_key;
	devreg_write_unlock_devices();
}

static struct devreg_extension* _find_extension(const char *name)
{					      
	
	struct list_head *pos;
	struct devreg_phydev *dev;
	struct devreg_extension_instance *ex;

	list_for_each(pos, &devreg_devices) {
		dev = list_entry(pos, struct devreg_phydev, phydevs);
		
		VECTOR_FOR_EACH(&dev->extensions,
				ex,
				struct devreg_extension_instance) {
			if (!strcmp(name, ex->extension->name))
				return ex->extension;
		}
	}
	return NULL;
}

static struct devreg_extension* _get_extension(const char *name,
					       int (*register_files) (
						       struct proc_dir_entry *dir, 
						       void *data),
					       void *private_data,
					       proc_ov_dir_enter_t enter_cb,
					       proc_ov_dir_leave_t leave_cb)
{
	struct devreg_extension *r;
	r = _find_extension(name);
	if (r)
		return r;
	return devreg_create_extension(name, 
				       register_files, 
				       private_data, 
				       enter_cb, 
				       leave_cb);
}

static void _add_extension(struct devreg_phydev *dev,
			   struct devreg_extension* dx,
			   void *context)
{
	struct devreg_extension_instance ei;
	ei.extension = dx;
	ei.context = context;
	
	if (VECTOR_ADD(&dev->extensions, &ei, 
		       struct devreg_extension_instance, GFP_KERNEL)) 
		dx->rc++;
	else {
		printk(KERN_ERR "devreg: failure registering extension: "
		       "out of memory\n");
		if (!dx->rc)
			devreg_free_extension(dx);
	}
}

int devreg_compare_extension_instance(void *elem, void *priv) 
{
	struct devreg_extension_instance* ei = 
		(struct devreg_extension_instance*) elem;
	struct devreg_extension* dx = (struct devreg_extension*) priv;

	return (ei->extension == dx) ? 0 : 1;
}

static void _remove_extension(struct devreg_phydev *dev,
			      struct devreg_extension* dx)
{
	VECTOR_REMOVE_FUNC_REORDER(&dev->extensions, 
				   struct devreg_extension_instance, 
				   devreg_compare_extension_instance, 
				   dx, 
				   GFP_KERNEL);
	dx->rc--;
	if (!dx->rc)
		devreg_free_extension(dx);
}

void devreg_register_extension(void *key,
			       const char *name, 
			       int (*register_files) (struct proc_dir_entry *dir, 
						      void *data),
			       void *private_data,
			       proc_ov_dir_enter_t enter_cb,
			       proc_ov_dir_leave_t leave_cb,
			       void *context)
{
	struct devreg_phydev *dev;
	struct devreg_extension *dx;

        if (!key) 
		return;

	devreg_write_lock_devices();
	
	dev = _find_physical_device(key);
	if (!dev) {
		devreg_write_unlock_devices();
		printk(KERN_ERR "devreg: failure registering extension: "
		       "invalid key\n");
		return;
	}
	
	dx = _get_extension(name, register_files, private_data, enter_cb, leave_cb);
	if (!dx) {
		devreg_write_unlock_devices();
		printk(KERN_ERR "devreg: failure registering extension: "
		       "insufficient memory or proc registration failed\n");
		return;
	}
	
	_add_extension(dev, dx, context);
	
	devreg_write_unlock_devices();
}

void devreg_unregister_extension(void *key, 
				 const char *name)
{
	struct devreg_phydev *dev;
	struct devreg_extension *ex;

	if (!key)
		return;

	devreg_write_lock_devices();
	
	dev = _find_physical_device(key);
	if (!dev) {
		devreg_write_unlock_devices();
		printk(KERN_ERR "devreg: failure unregistering extension: "
		       "invalid key\n");
		return;
	}

	ex = _find_extension(name);
	if (!ex) {
		devreg_write_unlock_devices();
		printk(KERN_ERR "devreg: failure unregistering extension: "
		       "unknown extension\n");
		return;
	}

	_remove_extension(dev, ex);
	
	devreg_write_unlock_devices();
}

void devreg_register_devfs_handle(void *key,
				  devfs_handle_t handle)
{
	struct devreg_phydev *dev;

        if ((!key) || (!handle)) 
		return;

	devreg_write_lock_devices();
	
	dev = _find_physical_device(key);
	if (!dev) {
		devreg_write_unlock_devices();
		printk(KERN_ERR "devreg: failure registering device file: "
		       "invalid key\n");
		return;
	}

	if(!VECTOR_ADD(&dev->devfiles, &handle, devfs_handle_t, GFP_KERNEL)) {
		printk(KERN_ERR "devreg: failure registering device file: "
		       "out of memory\n");
	}

	devreg_write_unlock_devices();
}

void devreg_unregister_devfs_handle(void *key,
				    devfs_handle_t handle) 
{
	struct devreg_phydev *dev;

	if ((!key) || (!handle))
		return;

	devreg_write_lock_devices();
	
	dev = _find_physical_device(key);
	if (!dev) {
		devreg_write_unlock_devices();
		printk(KERN_ERR 
		       "devreg: failure unregistering device file: "
		       "invalid key\n");
		return;
	}
	VECTOR_REMOVE_ELEMENT_REORDER(&dev->devfiles, 
				      &handle, 
				      devfs_handle_t, 
				      GFP_KERNEL);

	devreg_write_unlock_devices();
}

void __init devreg_init() 
{
	init_rwsem(&devreg_devices_lock);
	devreg_change_count = 0;
	devreg_proc_init();
}

EXPORT_SYMBOL(devreg_register_physical_device);
EXPORT_SYMBOL(devreg_set_device_type);
EXPORT_SYMBOL(devreg_set_device_serial_number);
EXPORT_SYMBOL(devreg_unregister_physical_device);
EXPORT_SYMBOL(devreg_get_physical_device);
EXPORT_SYMBOL(devreg_register_extension);
EXPORT_SYMBOL(devreg_unregister_extension);


