#ifndef LINUX_DEVREG_T_H
#define LINUX_DEVREG_T_H

#include <linux/devreg.h>
#include <linux/wait.h>
#include <linux/list.h>
#include <linux/vector.h>
#include <linux/kdev_t.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/proc_fs.h>

#include <asm/semaphore.h>

/*
 * Contains the information needed for a extension
 */
struct devreg_extension {
	const char *name;
	struct proc_dir_entry *dir;
	int rc; /* reference count */
	proc_ov_dir_enter_t *enter_cb;
	proc_ov_dir_leave_t *leave_cb;
};

/*
 * Per-instance data for each devreg_pydev and extension
 */
struct devreg_extension_instance {
	struct devreg_extension *extension;
	void *context;
};

/*
 * Describes a physical device
 */
struct devreg_phydev {
	struct list_head phydevs; /* node in phydev_list */
	void *parent_key;      
	void *superdevice_key; 
	const char *bus;       /* name/id of bus */
	const char *type;
	const char *deviceidwosn;   /* device id without serialnum 
                                      (but with the last '/')     */  
	const char *deviceidsernum; /* device id serialnum         */
	const char *model;     
	const char *vendor;         /* NULL if unknown */ 
	const char *location_desc;  /* location in text form */  
	enum devreg_location location;

	void *key;  /* key to find devices, never NULL, bus-specific */

	struct vector extensions; /* list of struct devreg_extension_instance */

	struct vector devfiles;  /* list of devfs_handle_t */
};

/*
 * Used to find extensions in a devreg_phydev.extensions.
 */
int devreg_compare_extension_instance(void *elem, void *priv);

/*
 * Register the devfs_handle_t for the device with the given key.
 * Locks devreg.
 */
void devreg_register_devfs_handle(void *key,
				  devfs_handle_t handle);

/*
 * Unregister the devfs_handle_t of the device with the given key.
 * Locks devreg.
 */
void devreg_unregister_devfs_handle(void *key,
				    devfs_handle_t handle);

/*
 * Creates a new extension in the proc file system
 */
struct devreg_extension* devreg_create_extension(const char *name,
						 int (*register_files) (
							 struct proc_dir_entry *dir, 
							 void *data),
						 void *private_data,
						 proc_ov_dir_enter_t enter_cb,
						 proc_ov_dir_leave_t leave_cb);

/*
 * Frees a extension
 */
void devreg_free_extension(struct devreg_extension *ex);

void devreg_write_lock_devices(void);
void devreg_write_unlock_devices(void);
#define devreg_read_lock_devices() down_read(&devreg_devices_lock)
#define devreg_read_unlock_devices() up_read(&devreg_devices_lock)

/*
 * List of struct devreg_phydev with all devices.
 * Locked by devreg_devices_lock, use devreg_[read|write]_lock_devices().
 */
extern struct list_head devreg_devices;

/*
 * Change count to invalidate user access when devices change.
 * Locked by devreg_devices_lock, use devreg_[read|write]_lock_devices().
 */
extern volatile int devreg_change_count;

/* 
 * Lock for devices.
 */
extern struct rw_semaphore devreg_devices_lock; 

/*
 * A wait queue that will waked up after each change (and after
 * the lock has been released).
 */
extern wait_queue_head_t devreg_change_queue;

#endif
