/**
 * proc.c
 * 
 * (C) Copyright Tim Jansen 2001 <tim@tjansen.de>
 *
 * Registers the /proc/devreg files, provides their callbacks.
 *
 */

#include "devreg_t.h"
#include "proc.h"
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <asm/uaccess.h>

#define MAX_DEVS  128
#define MAX_NODES   8
#define PATHLEN 128 /* is there a kernel constant for the maximum path length? */

static struct proc_dir_entry *droot, *dperdevice;


static struct devreg_extension_instance*
get_extension_instance_ptr(struct devreg_phydev *dev, 
			   struct devreg_extension *ex) {
	int index;

	index = VECTOR_FIND(&dev->extensions, 
			    struct devreg_extension_instance,
			    devreg_compare_extension_instance, 
			    ex);
	if (index < 0)
		return NULL;
	return VECTOR_GET_ELEMENTPTR_AT(&dev->extensions,
					index, 
					struct devreg_extension_instance);
}

static int extension_num(void *context, 
			 void *data) 
{
	struct devreg_phydev *dev = (struct devreg_phydev*) context;
	struct devreg_extension *ex = (struct devreg_extension*) data;

	if (! (dev && ex))
		return 0;

	return get_extension_instance_ptr(dev, ex) ? 1 : 0;
}
static void *extension_context(int index, 
			       void *context, 
			       void *data) 
{
	struct devreg_extension_instance *eip;
	struct devreg_phydev *dev = (struct devreg_phydev*) context;
	struct devreg_extension *ex = (struct devreg_extension*) data;

	if (!(dev && ex))
		return NULL;

	eip = get_extension_instance_ptr(dev, ex);
	if (!eip)
		return NULL;

	if (ex->enter_cb)
		return ex->enter_cb(eip->context, NULL);

	return eip->context;
}
static int extension_name(int index, 
			  void *context, 
			  void *data,
			  char *name_buf)
{
	struct devreg_extension *ex = (struct devreg_extension*) data;

	if (!ex)
		return -ENOENT;
	return sprintf(name_buf, "%s", ex->name);
}
static void extension_leave(void *old_context,
			    void *new_context,
			    void *data) 
{
	struct devreg_extension *ex = (struct devreg_extension*) data;
	if (ex && ex->leave_cb) 
		ex->leave_cb(ex->enter_cb ? old_context :  NULL, 
			     new_context, 
			     NULL);	
}

static int write_change_count(int *value,
			      void *context,
			      void *data)
{
	*value = devreg_change_count;
	return 0;
}

static int changecount_open(struct inode *i, struct file *f)
{
	f->private_data = (void*) devreg_change_count;
	return 0;
}

#define CHANGECOUNT_BUFLEN 16
#define PRINT_CHANGECOUNT(b) sprintf(b, "%d\n", devreg_change_count);

static ssize_t changecount_read(struct file *f, 
				char *dest, 
				size_t s, 
				loff_t *o)
{
	char buf[CHANGECOUNT_BUFLEN];
	int len;
	int size = s;

	if ((!*o) && s)
		f->private_data = (void*) devreg_change_count;

	len = PRINT_CHANGECOUNT(buf);
	
	if ((size+*o) > len) 
		size = len - *o;
	if (size <= 0)
		return 0;

	copy_to_user(dest, buf + *o, size);
	*o += size;
	return size;
}

static loff_t changecount_llseek(struct file *f, loff_t o, int type)
{
    switch (type) {
    case 0:
	    if (o < 0)
		    return -EINVAL;    
	    f->f_pos = o;
	    return o;
    case 1:
	    if ((o + f->f_pos) < 0)
		    return -EINVAL;    
	    f->f_pos += o;
	    return f->f_pos;
    default:
	    return(-EINVAL);
    }
}

static unsigned int changecount_poll (struct file *f, 
				      struct poll_table_struct *pts)
{
	char buf[CHANGECOUNT_BUFLEN];
	int len;
	unsigned int flags = 0;

	len = PRINT_CHANGECOUNT(buf);

	poll_wait(f, &devreg_change_queue, pts);

	if (len > f->f_pos)
		flags |= POLLIN | POLLRDNORM;

	if (devreg_change_count != (int)f->private_data)
		flags |= POLLERR;

	return flags;
}

static int perdevice_num(void *context, 
			 void *data) 
{
	int n = 0;
	struct list_head *pos;

	list_for_each(pos, &devreg_devices) {
		n++;
	}	
	return n;
}
static void *perdevice_context(int index, 
			       void *context, 
			       void *data) 
{
	int n = 0;
	struct list_head *pos;

	list_for_each(pos, &devreg_devices) {
		if (index != n) { 
			n++;
			continue;
		}
		devreg_read_lock_devices();
		return list_entry(pos, struct devreg_phydev, phydevs);
	}	
	return NULL;
}
static void perdevice_leave(void *old_context,
			    void *new_context,
			    void *data) 
{
	if (new_context)
		devreg_read_unlock_devices();
}


static int write_charfield(char *buf,
			   void *context,
			   void *data) 
{
	char **cp;
	if (!context)
		return -ENOENT;
	cp = (char**) (context + (int) data);
	return sprintf(buf, "%s", *cp ? *cp : "");
}

static int write_deviceid(char *buf,
			  void *context,
			  void *data) 
{
	struct devreg_phydev *dev = (struct devreg_phydev*) context;
	if (!context)
		return -ENOENT;
	return sprintf(buf, 
		       "%s/%s", 
		       dev->deviceidwosn, 
		       dev->deviceidsernum ? dev->deviceidsernum : "");
}

static const char *location_table[] = { 
	"internal", 
	"external", 
	"virtual", 
	"unknown",
	NULL
};

static int write_location(int *value,
			  void *context,
			  void *data) 
{
	struct devreg_phydev *dev = (struct devreg_phydev*) context;
	if (!context)
		return -ENOENT;
	*value = (int) dev->location;
	return 0;
}

static int write_key_deviceid(char *buf,
			      void *context,
			      void *data) 
{
	void **keyptr;
	void *key;
	struct list_head *pos;
	struct devreg_phydev *dp = NULL;

	if (!context)
		return -ENOENT;
	keyptr = (void**) (context + (int) data);
	key = *keyptr;

	list_for_each(pos, &devreg_devices) {
		dp = list_entry(pos, struct devreg_phydev, phydevs);
		if (dp->key == key)
			return write_deviceid(buf, dp, NULL);
	}
	buf = '\0';
	return 0;
}

static int pernode_num(void *context, 
		       void *data) 
{
	struct devreg_phydev *dev = (struct devreg_phydev*) context;
	return VECTOR_GET_SIZE(&dev->devfiles);
}
static void *pernode_context(int index, 
			     void *context, 
			     void *data) 
{
	struct devreg_phydev *dev = (struct devreg_phydev*) context;
	struct vector *v;

	if (!context)
		return NULL;
	
	v = &dev->devfiles;

	if ((index < 0) || (index >= VECTOR_GET_SIZE(v)))
		return NULL;

	return VECTOR_GET_ELEMENTPTR_AT(v, index, devfs_handle_t);
}

static int write_node_name(char *buf,
			   void *context,
			   void *data) 
{
	char pathbuf[PATHLEN];
	devfs_handle_t h;
	int offset;

	if (!context)
		return -ENOENT;
	h = *((devfs_handle_t*) context);

        offset = devfs_generate_path(h, pathbuf, PATHLEN);
        if (offset < 0) {
                printk(KERN_ERR "devreg: Cannot generate devfile path");
                return -ENOENT;
        }
 
        return sprintf(buf, "%s", pathbuf+offset);
}

static int write_node_interface(char *buf,
				void *context,
				void *data) 
{
	const char *interface;
	devfs_handle_t h;

	if (!context)
		return -ENOENT;
	h = *((devfs_handle_t*) context);

	interface = devfs_get_interface(h);
        if (!interface)
                return -ENOENT;
 
        return sprintf(buf, "%s", interface);
}

struct devreg_extension* devreg_create_extension(const char *name,
						 int (*register_files) (
							 struct proc_dir_entry *dir, 
							 void *data),
						 void *private_data,
						 proc_ov_dir_enter_t enter_cb,
						 proc_ov_dir_leave_t leave_cb)
{
	struct devreg_extension *ex;
	ex = kmalloc(sizeof(struct devreg_extension), GFP_KERNEL);
	if (!ex)
		return NULL;

	ex->dir = proc_ov_register_dynamic_dir(dperdevice, 
					       1, 
					       extension_num, 
					       extension_name,
					       extension_context, 
					       extension_leave);
	if (!ex->dir) {
		kfree(ex);
		return NULL;
	}
	
	if (register_files(ex->dir, private_data) < 0) {
		devreg_free_extension(ex);
		return NULL;
	}

	ex->name = name;
	ex->rc = 0;
	ex->dir->data = ex;
	ex->leave_cb = leave_cb;
	ex->enter_cb = enter_cb;
	return ex;
}

void devreg_free_extension(struct devreg_extension *ex) 
{
	proc_ov_remove(ex->dir);
	kfree(ex);
}


void __init devreg_proc_init()
{
	struct proc_dir_entry *change_count;
	struct proc_dir_entry *busname, *deviceid, *model, *vendor;
	struct proc_dir_entry *type, *location_desc, *location;
	struct proc_dir_entry *superdevice, *parent, *nodes;
	struct proc_dir_entry *dpernode, *node_name, *node_interface;	
	struct file_operations *changecount_fp;

	#define CHECK(x) if (!x) {printk(KERN_ERR "devreg: cannot register " \
					 "proc files\n"); \
                             proc_ov_remove(droot); \
                             return; }
	#define OFFS(member) (&((struct devreg_phydev *)0)->member)
 

	droot = proc_ov_register_dir(NULL, "devreg", NULL, NULL);
	CHECK(droot);

	change_count = create_proc_entry("changecount", 0, droot);
	if (!change_count) {
		printk(KERN_ERR "devreg: cannot register "
		       "changecount file\n");
		remove_proc_entry("changecount", droot);
		return; 
	}
	changecount_fp = kmalloc(sizeof(struct file_operations), GFP_KERNEL);
	if (!changecount_fp) {
		printk(KERN_ERR "Unsufficient memory for the device"
		       " registry\n");
		return;
	}
	*changecount_fp = *change_count->proc_fops;
	changecount_fp->open = changecount_open;
	changecount_fp->read = changecount_read;
	changecount_fp->poll = changecount_poll;
	changecount_fp->llseek = changecount_llseek;
	change_count->proc_fops = changecount_fp;

	dperdevice = proc_ov_register_dynamic_dir(droot, 
						  MAX_DEVS,
						  perdevice_num,
						  NULL,
						  perdevice_context,
						  perdevice_leave);
	CHECK(dperdevice);

	busname = proc_ov_register_string(dperdevice, "bus", write_charfield);
	CHECK(busname);
	busname->data = OFFS(bus);

	deviceid = proc_ov_register_string(dperdevice, "deviceid", 
					   write_deviceid);
	CHECK(deviceid);

	model = proc_ov_register_string(dperdevice, "model", write_charfield);
	CHECK(model);
	model->data = OFFS(model);

	vendor = proc_ov_register_string(dperdevice, "vendor", 
					 write_charfield);
	CHECK(vendor);
	vendor->data = OFFS(vendor);

	type = proc_ov_register_string(dperdevice, "type", write_charfield);
	CHECK(type);
	type->data = OFFS(type);

	location_desc = proc_ov_register_string(dperdevice, "locationdesc", 
						write_charfield);
	CHECK(location_desc);
	location_desc->data = OFFS(location_desc);
	
	location = proc_ov_register_enum(dperdevice, "location", 
					 location_table,
					 write_location);
	CHECK(location);

	parent = proc_ov_register_string(dperdevice, "parent", 
					 write_key_deviceid);
	CHECK(parent);
	parent->data = OFFS(parent_key);

	superdevice = proc_ov_register_string(dperdevice, "superdevice", 
					      write_key_deviceid);
	CHECK(superdevice);
	superdevice->data = OFFS(superdevice_key);

	nodes = proc_ov_register_dir(dperdevice, "devicenodes", NULL, NULL);
	CHECK(nodes);

	dpernode = proc_ov_register_dynamic_dir(nodes, 
						MAX_NODES,
						pernode_num,
						NULL,
						pernode_context,
						NULL);
	CHECK(dpernode);

	node_name = proc_ov_register_string(dpernode, "name", 
					    write_node_name);
	CHECK(node_name);

	node_interface = proc_ov_register_string(dpernode, "interface", 
						 write_node_interface);
	CHECK(node_interface);
	
}

