/*
 * proc/fs/generic.c --- generic routines for the proc-fs
 *
 * This file contains generic proc-fs routines for handling
 * directories and files.
 * 
 * Copyright (C) 1991, 1992 Linus Torvalds.
 * Copyright (C) 1997 Theodore Ts'o
 */

#include <asm/uaccess.h>

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/proc_fs.h>
#include <linux/stat.h>
#include <linux/list.h>
#include <linux/init.h>
#define __NO_VERSION__
#include <linux/module.h>
#include <asm/bitops.h>

/* size for context stacks */
#define CONTEXT_STACK_SIZE (PROC_OV_MAX_CONTEXT_DEPTH+1)

struct inode_range {
	struct list_head free_inodes;
	unsigned int first, end;
};
static LIST_HEAD(free_inodes);


static ssize_t proc_file_read(struct file * file, char * buf,
			      size_t nbytes, loff_t *ppos);
static ssize_t proc_file_write(struct file * file, const char * buffer,
			       size_t count, loff_t *ppos);
static loff_t proc_file_lseek(struct file *, loff_t, int);

int proc_match(int len, const char *name,struct proc_dir_entry * de)
{
	if (!de || !de->low_ino)
		return 0;
	if (de->namelen != len)
		return 0;
	return !memcmp(name, de->name, len);
}

static void *enter_context_pde(const struct dentry *dentry,
	                       struct proc_dir_entry *npde,
			       void **context_stack) 
{
	void *c;
	struct proc_dir_entry *pde;

	if (!dentry) {
		*context_stack = NULL;
		return NULL;
	}
	
	if (npde)
		pde = npde;
	else
		pde = dentry->d_inode->u.proc_ov_i.entry;
	
	if ((!pde) || 
	    ((pde->type != PROC_OV_DIR) &&
	     (pde->type != PROC_OV_DYNAMIC_DIR))) {
		*context_stack = NULL;
		return NULL;
	}

	c = enter_context_pde(dentry->d_parent, NULL, context_stack + 1);

	if (pde->type == PROC_OV_DIR) {
		struct proc_ov_dir *ov = (struct proc_ov_dir *) pde;
		if (ov->enter) 
			c = ov->enter(c, 
				      pde->data);
	}
	else {
		struct proc_ov_dynamic_dir *ov = 
			(struct proc_ov_dynamic_dir *) pde;
                if (ov->get_context)
			c = ov->get_context(dentry->d_inode->u.proc_ov_i.dir_index, 
					    c,
					    pde->data);
	}

	*context_stack = c;
	return c;
}

static void *enter_context(const struct dentry *dentry,
			   void **context_stack) 
{
	enter_context_pde(dentry, NULL, context_stack);
}

static void leave_context_pde(const struct dentry *dentry,
			      struct proc_dir_entry *npde,
			      void **context_stack) 
{
	struct proc_dir_entry *pde;

	if (!dentry)
		return;

	if (npde)
		pde = npde;
	else
		pde = dentry->d_inode->u.proc_ov_i.entry;

	if ((!pde) || 
	    ((pde->type != PROC_OV_DIR) &&
	     (pde->type != PROC_OV_DYNAMIC_DIR)))
		return;

	if (pde->type == PROC_OV_DIR) {
		struct proc_ov_dir *ov = (struct proc_ov_dir *) pde;
		if (ov->leave)
			ov->leave(context_stack[1], 
				  *context_stack,
				  pde->data);
	}
	else {
		struct proc_ov_dynamic_dir *ov = 
			(struct proc_ov_dynamic_dir *) pde;
		if (ov->leave)
			ov->leave(context_stack[1], 
				  *context_stack,
				  pde->data);
	}

	leave_context(dentry->d_parent, context_stack+1);
}

static void *leave_context(const struct dentry *dentry,
			   void **context_stack) 
{
	leave_context_pde(dentry, NULL, context_stack);
}



static struct file_operations proc_file_operations = {
	llseek:		proc_file_lseek,
	read:		proc_file_read,
	write:		proc_file_write,
};

#ifndef MIN
#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#endif

static ssize_t
proc_file_ov_read(struct file *file, 
		  char *page, 
		  struct proc_dir_entry *dp)
{
	void *context_stack[CONTEXT_STACK_SIZE];
	void *context;
	struct proc_ov_value_enum *ove;
	ssize_t n = 0;
	int r = 0;
	int value;
	unsigned long ulvalue;

	context = enter_context(file->f_dentry->d_parent, 
				context_stack);
	switch(dp->type) {
	case PROC_OV_STRING:
		r = ((struct proc_ov_value_string*)dp)
			->get_string(page, context, dp->data);
		if (r >= 0) {
			n = r + 1;
			page[n-1] = '\n';
			page[n] = 0;					
		}
		break;
	case PROC_OV_INT:
		r = ((struct proc_ov_value_int*)dp)
			->get_int(&value, context, dp->data);
		if (r >= 0)
			n = sprintf(page, "%d\n", value);
		break;
	case PROC_OV_ULONG:
		r = ((struct proc_ov_value_ulong*)dp)
			->get_ulong(&ulvalue, context, dp->data);
		if (r >= 0)
			n = sprintf(page, "%0lx\n", ulvalue);
		break;
	case PROC_OV_ENUM:
		ove = (struct proc_ov_value_enum*)dp;
		r = ove->get_enum(&value, context, dp->data);
		if (r >= 0) {
			strcpy(page, ove->enum_values[value]);
			n = strlen(page) + 1;
			page[n-1] = '\n';
			page[n] = 0;
		}
		break;
	default:
		r = -EISDIR;
	}
	leave_context(file->f_dentry->d_parent, context_stack);
	if (r < 0)
		return r;
	return n;
}

static ssize_t
proc_file_read(struct file * file, char * buf, size_t nbytes, loff_t *ppos)
{
	struct inode * inode = file->f_dentry->d_inode;
	char 	*page;
	ssize_t	retval=0;
	int	eof=0;
	ssize_t	n, count;
	char	*start;
	struct proc_dir_entry * dp;

	dp = inode->u.proc_ov_i.entry;
	if (!(page = (char*) __get_free_page(GFP_KERNEL)))
		return -ENOMEM;
	n = 0;

	while ((nbytes > 0) && !eof)
	{
		count = MIN(PROC_BLOCK_SIZE, nbytes);

		start = NULL;
		if (dp->get_info) {
			/*
			 * Handle backwards compatibility with the old net
			 * routines.
			 */
			n = dp->get_info(page, &start, *ppos, count);
			if (n < count)
				eof = 1;
		} else if (dp->read_proc) {
			n = dp->read_proc(page, &start, *ppos,
					  count, &eof, dp->data);
		} else if (dp->type != PROC_STATIC_ENTRY) {
			n = proc_file_ov_read(file, page, dp);
		} else
			break;

		if (n < 0) {
			if (retval == 0)
				retval = n;
			break;
		}
		if (!start) {
			/*
			 * For proc files that are less than 4k
			 */
			start = page + *ppos;
			n -= *ppos;
			if (n <= 0)
				break;
			if (n > count)
				n = count;
		}
		if (n == 0)
			break;	/* End of file */
		
		/* This is a hack to allow mangling of file pos independent
 		 * of actual bytes read.  Simply place the data at page,
 		 * return the bytes, and set `start' to the desired offset
 		 * as an unsigned int. - Paul.Russell@rustcorp.com.au
		 */
 		n -= copy_to_user(buf, start < page ? page : start, n);
		if (n == 0) {
			if (retval == 0)
				retval = -EFAULT;
			break;
		}

		*ppos += start < page ? (long)start : n; /* Move down the file */
		nbytes -= n;
		buf += n;
		retval += n;
	}
	free_page((unsigned long) page);
	return retval;
}

static ssize_t
proc_file_write(struct file * file, const char * buffer,
		size_t count, loff_t *ppos)
{
	struct inode *inode = file->f_dentry->d_inode;
	struct proc_dir_entry * dp;
	
	dp = inode->u.proc_ov_i.entry;

	if (!dp->write_proc)
		return -EIO;

	/* FIXME: does this routine need ppos?  probably... */
	return dp->write_proc(file, buffer, count, dp->data);
}


static loff_t
proc_file_lseek(struct file * file, loff_t offset, int orig)
{
    switch (orig) {
    case 0:
	if (offset < 0)
	    return -EINVAL;    
	file->f_pos = offset;
	return(file->f_pos);
    case 1:
	if (offset + file->f_pos < 0)
	    return -EINVAL;    
	file->f_pos += offset;
	return(file->f_pos);
    case 2:
	return(-EINVAL);
    default:
	return(-EINVAL);
    }
}

/*
 * This function parses a name such as "tty/driver/serial", and
 * returns the struct proc_dir_entry for "/proc/tty/driver", and
 * returns "serial" in residual.
 * Does not find dynamic entries (type == PROC_OV_DYNAMIC_DIR)
 */
static int xlate_proc_name(const char *name,
			   struct proc_dir_entry **ret, const char **residual)
{
	const char     		*cp = name, *next;
	struct proc_dir_entry	*de;
	int			len;

	de = &proc_root;
	while (1) {
		next = strchr(cp, '/');
		if (!next)
			break;

		len = next - cp;
		for (de = de->subdir; de ; de = de->next) {
			if (proc_match(len, cp, de))
				break;
		}
		if (!de)
			return -ENOENT;
		cp += len + 1;
	}
	*residual = cp;
	*ret = de;
	return 0;
}



static unsigned int allocate_inode_range(unsigned int n) 
{
	struct list_head *pos;
	struct list_head h;
	struct inode_range *r, *prev, *next;
	unsigned i = 0;

	list_for_each(pos, &free_inodes) {
		r = list_entry(pos, struct inode_range, free_inodes);

		if ((r->end - r->first) < n)
			continue;
		i = r->first;
		r->first += n;
		if (r->first != r->end) 
			break;

		h = *pos;
		list_del(pos);
		kfree(r);

		if ((h.prev == &free_inodes) ||
		    (h.next == &free_inodes))
			break;
		prev = list_entry(h.prev, struct inode_range, free_inodes);
		next = list_entry(h.next, struct inode_range, free_inodes); 
		if (prev->end != next->first)
			break;
		next->first = prev->first;
		list_del(h.prev);
		kfree(prev);
		break;
	}

	if (!i)
		printk(KERN_ERR "proc: out of inodes\n");
	return i;
}

static struct list_head *_prepend_inode_range(struct list_head *l, 
					      unsigned int ino, 
					      unsigned int n) {
	struct inode_range *r;

	r = kmalloc(sizeof(*r), GFP_KERNEL);
	if (!r)
		return NULL;
	r->first = ino;
	r->end = ino + n;
	list_add_tail(&r->free_inodes, l);
	return &r->free_inodes;
}
static void _join_inode_ranges(struct list_head *center) {
	struct inode_range *p, *c, *n;
	c = list_entry(center, struct inode_range, free_inodes);

	if (center->prev != &free_inodes) {
		p = list_entry(center->prev, struct inode_range, free_inodes);
		if (p->end == c->first) {
			c->first = p->first;
			list_del(center->prev);
			kfree(p);
		}
	}

	if (center->next != &free_inodes) {
		n = list_entry(center->next, struct inode_range, free_inodes);
		if (c->end == n->first) {
			c->end = n->end;
			list_del(center->next);
			kfree(n);
		}
	}	
}
static int free_inode_range(unsigned int ino, unsigned int n) 
{
	struct list_head *pos;
	struct list_head *c;

	list_for_each(pos, &free_inodes) {
		struct inode_range *r;
		r = list_entry(pos, struct inode_range, free_inodes);
		if (ino < r->first)
			break;
	}
	
	c = _prepend_inode_range(pos, ino, n);
	if(!c)
		return -1;

	_join_inode_ranges(c);
	return 0;
}


static int proc_readlink(struct dentry *dentry, char *buffer, int buflen)
{
	char *s=(dentry->d_inode->u.proc_ov_i.entry)->data;
	return vfs_readlink(dentry, buffer, buflen, s);
}

static int proc_follow_link(struct dentry *dentry, struct nameidata *nd)
{
	char *s=(dentry->d_inode->u.proc_ov_i.entry)->data;
	return vfs_follow_link(nd, s);
}

static struct inode_operations proc_link_inode_operations = {
	readlink:	proc_readlink,
	follow_link:	proc_follow_link,
};

/*
 * As some entries in /proc are volatile, we want to 
 * get rid of unused dentries.  This could be made 
 * smarter: we could keep a "volatile" flag in the 
 * inode to indicate which ones to keep.
 */
static int proc_delete_dentry(struct dentry * dentry)
{
	return 1;
}

static struct dentry_operations proc_dentry_operations =
{
	d_delete:	proc_delete_dentry,
};

static unsigned int get_inode_space(struct proc_dir_entry *p) 
{
	if ((!p) || (p->type == PROC_STATIC_ENTRY))
		return 1;
	if (p->type == PROC_OV_DIR)
		return ((struct proc_ov_dir*)p)->ino_space;
	if (p->type == PROC_OV_DYNAMIC_DIR)
		return ((struct proc_ov_dynamic_dir*)p)->ino_space;
      	return get_inode_space(p->parent);
}

static unsigned int get_ino_index(struct inode *parent_inode, 
				  struct proc_dir_entry *de,
				  int index) 
{
	unsigned int ino;
	if (de->type == PROC_STATIC_ENTRY)
		return 0;
	if ((!de->parent) || (de->parent->type == PROC_STATIC_ENTRY))
		ino = 0;
	else
		ino = parent_inode->u.proc_ov_i.inode_index;

	if (de->type == PROC_OV_DYNAMIC_DIR) {
		ino *= ((struct proc_ov_dynamic_dir *)de)->max_size; 
		ino += index;
	}
	return ino;
}

static int get_dyndir_links(struct proc_dir_entry *de,
			    struct dentry *dentry)
{
	struct proc_ov_dynamic_dir *dd;
	int nlinks = 0;
	void *context_stack[CONTEXT_STACK_SIZE];
	void *context;

	dd = de->dyndir;
	
	context = enter_context_pde(dentry, de, context_stack);
	while (dd) {
		nlinks += dd->get_num(context, dd->super.data);
		dd = (struct proc_ov_dynamic_dir*)dd->super.next;
	}
	leave_context_pde(dentry, de, context_stack);
	
	return nlinks;
}

static struct inode *_lookup_static_inode(struct proc_dir_entry *parent,
					  struct inode *dir,
					  struct dentry *dentry,
					  int *error) 
{
	struct proc_dir_entry *de;
	struct inode *inode = NULL;
	unsigned int ino;

	for (de = parent->subdir; de ; de = de->next) {
		if (!de || !de->low_ino)
			continue;
		if (de->namelen != dentry->d_name.len)
			continue;
		if (!memcmp(dentry->d_name.name, de->name, de->namelen)) {
			*error = -EINVAL;
			ino = get_ino_index(dir, de, 0);

			inode = proc_get_inode(dir->i_sb, 
					       ino + de->low_ino, 
					       de);
			inode->u.proc_ov_i.inode_index = ino;
			inode->u.proc_ov_i.dir_index = -1;
			if (de->dyndir)
				inode->i_nlink += get_dyndir_links(de, dentry);

			break;
		}
	}
	return inode;
}

static int index_number_name(int index, 
			     void *context, 
			     void *data,
			     char *name_buf)
{
	return sprintf(name_buf, "%d", index);
}

static int _match_regular_name(struct proc_ov_dynamic_dir *ov,
			       void *context, 
			       int n,
			       struct dentry *dentry)
{
	int nlen, i;
	char namebuf[NAME_MAX];
	for (i = 0; i < n; i++) {
		nlen = ov->get_name(i, 
				    context, 
				    ov->super.data,
				    namebuf);
		if ((nlen == dentry->d_name.len) &&
		    !strcmp(dentry->d_name.name, namebuf)) 
			return i;
	}
	return -1;
}
static int _match_number_name(int n, struct dentry *dentry)
{
	int r = 0;
	int l = dentry->d_name.len;
	const char *name = dentry->d_name.name;
	char c;

	while (l > 0) {
		c = *(name++);
		if ((c < '0') || (c > '9'))
			return -1;
		if ((c == '0') && (r == 0) && (*name != '\0'))
			return -1;
		r *= 10;
		r += c - '0';
		l--;
	}

	if (n > r)
		return r;
	return -1;
}
static struct inode *_lookup_dynamic_inode(struct proc_dir_entry *de,
					   struct inode *dir,
					   struct dentry *dentry,
					   int *error) 
{
	struct proc_ov_dynamic_dir *ov;
	struct inode *inode = NULL;
	unsigned int ino;
	int n, i;
	void *context_stack[CONTEXT_STACK_SIZE];
	void *context;

	ov = de->dyndir;
	if (!ov)
		return NULL;

	context = enter_context(dentry->d_parent, context_stack);
	while (ov) {
		n = ov->get_num(context, ov->super.data);
		if (n > ov->max_size) {
			printk(KERN_WARNING "proc: not enough inodes reserved "
			       "for dynamic dir (max_size=%d)\n", ov->max_size);
			n = ov->max_size;
		}

		if ((ov->get_name && ((i = _match_regular_name(ov, context, n, dentry)) >= 0)) ||
		    ((!ov->get_name) && ((i = _match_number_name(n, dentry)) >= 0))) {
			*error = -EINVAL;
			ino = get_ino_index(dir, &ov->super, i);
			inode = proc_get_inode(dir->i_sb, 
					       ino + ov->super.low_ino, 
					       &ov->super);
			inode->u.proc_ov_i.inode_index = ino;
			inode->u.proc_ov_i.dir_index = i;
			break;
		}
		ov = (struct proc_ov_dynamic_dir*) ov->super.next;
	}
	leave_context(dentry->d_parent, context_stack);
	return inode;
}

/*
 * Don't create negative dentries here, return -ENOENT by hand
 * instead.
 */
struct dentry *proc_lookup(struct inode * dir, struct dentry *dentry)
{
	struct inode *inode;
	struct proc_dir_entry *de;
	int error;

	error = -ENOENT;
	inode = NULL;
	de = dir->u.proc_ov_i.entry;
	if (!de) 
		return ERR_PTR(error);

	inode = _lookup_static_inode(de, dir, dentry, &error);
	if (!inode) 
		inode = _lookup_dynamic_inode(de, dir, dentry, &error);

	if (inode) {
		dentry->d_op = &proc_dentry_operations;
		d_add(dentry, inode);
		return NULL;
	}
	return ERR_PTR(error);
}

static int _read_static_dirs(struct proc_dir_entry *de, 
			     int *skip, 
			     struct file *filp,
			     void *dirent, 
			     filldir_t filldir)
{
	while (de && *skip) {
		de = de->next;
		(*skip)--;
	}
	
	while (de) {
		if (filldir(dirent, de->name, de->namelen, filp->f_pos,
			    get_ino_index(filp->f_dentry->d_inode, 
					  de, 
					  0) + de->low_ino, 
			    de->mode >> 12) < 0)
			return 1;
		filp->f_pos++;
		de = de->next;
	}
	return 0;
}

static int _read_dynamic_dirs(struct proc_ov_dynamic_dir *de, 
			      int *skip, 
			      struct file *filp,
			      void *dirent, 
			      filldir_t filldir)
{
	char namebuf[NAME_MAX];
	void *context_stack[CONTEXT_STACK_SIZE];
	int n, j, l;
	void *context;

	if (!de)
		return 0;

	context = enter_context(filp->f_dentry, 
				context_stack);
	while (de) {
		n = de->get_num(context, de->super.data);
		if (n > de->max_size) {
			printk(KERN_WARNING "proc: not enough inodes reserved "
			       "for dynamic dir (max_size=%d)\n", de->max_size);
			n = de->max_size;
		}
		for (j = *skip; j < n; j++) {
			if (de->get_name)
				l = de->get_name(j, context, de->super.data, namebuf);
			else
				l = sprintf(namebuf, "%d", j);

			if ( (l<=0) ||
			     (filldir(dirent, namebuf, l, 
				      filp->f_pos, 
				      get_ino_index(filp->f_dentry->d_inode, 
						    &de->super,
						    j) + de->super.low_ino, 
				      de->super.mode >> 12) < 0)) {
				leave_context(filp->f_dentry, context_stack);
				return 1;
			}
			filp->f_pos++;
		}
		*skip -= n;
		if (*skip < 0)
			*skip = 0;
		de = (struct proc_ov_dynamic_dir*) de->super.next;
	}
	leave_context(filp->f_dentry, context_stack);
	return 0;
}

/*
 * This returns non-zero if at EOF, so that the /proc
 * root directory can use this and check if it should
 * continue with the <pid> entries..
 *
 * Note that the VFS-layer doesn't care about the return
 * value of the readdir() call, as long as it's non-negative
 * for success..
 */
int proc_readdir(struct file * filp,
	void * dirent, filldir_t filldir)
{
	struct proc_dir_entry *de;
	unsigned int ino;
	int i;
	struct inode *inode = filp->f_dentry->d_inode;

	ino = inode->i_ino;
	de = inode->u.proc_ov_i.entry;
	if (!de)
		return -EINVAL;
	i = filp->f_pos;
	switch (i) {
		case 0:
			if (filldir(dirent, ".", 1, i, ino, DT_DIR) < 0)
				return 0;
			i++;
			filp->f_pos++;
			/* fall through */
		case 1:
			if (filldir(dirent, "..", 2, i,
				    filp->f_dentry->d_parent->d_inode->i_ino,
				    DT_DIR) < 0)
				return 0;
			i++;
			filp->f_pos++;
			/* fall through */
		default:
			i -= 2;

			if(_read_static_dirs(de->subdir, 
					     &i, 
					     filp, 
					     dirent, 
					     filldir))
				return 0;

			if (_read_dynamic_dirs(de->dyndir,
					       &i,
					       filp,
					       dirent,
					       filldir))
				return 0;
	}
	return 1;
}

/*
 * These are the generic /proc directory operations. They
 * use the in-memory "struct proc_dir_entry" tree to parse
 * the /proc directory.
 */
static struct file_operations proc_dir_operations = {
	read:			generic_read_dir,
	readdir:		proc_readdir,
};

/*
 * proc directories can do almost nothing..
 */
static struct inode_operations proc_dir_inode_operations = {
	lookup:		proc_lookup,
};

/* for onevalue.c */
int proc_init_dyn_dir(struct proc_dir_entry *ent, unsigned int ino_space)
{
	unsigned int ino;
	ent->proc_fops = &proc_dir_operations;
	ent->proc_iops = &proc_dir_inode_operations;	
	ino = allocate_inode_range(ino_space);
	ent->low_ino = ino;
	return ino != 0;
}

static int proc_register_ino(struct proc_dir_entry * dir, 
			     struct proc_dir_entry * dp,
			     unsigned int inos)
{
	unsigned int i = allocate_inode_range(inos);
	if (!i)
		return -EAGAIN;

	dp->low_ino = i;
	dp->next = dir->subdir;
	dp->parent = dir;
	dir->subdir = dp;
	if (S_ISDIR(dp->mode)) {
		if (dp->proc_iops == NULL) {
			dp->proc_fops = &proc_dir_operations;
			dp->proc_iops = &proc_dir_inode_operations;
		}
		dir->nlink++;
	} else if (S_ISLNK(dp->mode)) {
		if (dp->proc_iops == NULL)
			dp->proc_iops = &proc_link_inode_operations;
	} else if (S_ISREG(dp->mode)) {
		if (dp->proc_fops == NULL)
			dp->proc_fops = &proc_file_operations;
	}
	return 0;
}

static int proc_register(struct proc_dir_entry * dir, struct proc_dir_entry * dp) 
{
	return proc_register_ino(dir, dp, 1);
}

/*
 * Kill an inode that got unregistered..
 */
static void proc_kill_inodes(struct proc_dir_entry *de)
{
	struct list_head *p;
	struct super_block *sb = proc_mnt->mnt_sb;

	/*
	 * Actually it's a partial revoke().
	 */
	file_list_lock();
	for (p = sb->s_files.next; p != &sb->s_files; p = p->next) {
		struct file * filp = list_entry(p, struct file, f_list);
		struct dentry * dentry;
		struct inode * inode;

		dentry = filp->f_dentry;
		if (!dentry)
			continue;
		if (dentry->d_op != &proc_dentry_operations)
			continue;
		inode = dentry->d_inode;
		if (inode->u.proc_ov_i.entry != de)
			continue;
		fops_put(filp->f_op);
		filp->f_op = NULL;
	}
	file_list_unlock();
}

/*
 * Does not find dynamic entries (type == PROC_OV_DYNAMIC_DIR)
 */
static struct proc_dir_entry *proc_create(struct proc_dir_entry **parent,
					  const char *name,
					  mode_t mode,
					  nlink_t nlink,
					  unsigned int structlen)
{
	struct proc_dir_entry *ent = NULL;
	const char *fn = name;
	int len;

	/* make sure name is valid */
	if (!name || !strlen(name)) goto out;

	if (!(*parent) && xlate_proc_name(name, parent, &fn) != 0)
		goto out;
	len = strlen(fn);

	ent = kmalloc(structlen + len + 1, GFP_KERNEL);
	if (!ent)
		goto out;
	memset(ent, 0, structlen);
	memcpy(((char *) ent) + structlen, fn, len + 1);
	ent->name = ((char *) ent) + structlen;
	ent->namelen = len;
	ent->mode = mode;
	ent->nlink = nlink;
 out:
	return ent;
}

/*
 * Does not find dynamic entries (type == PROC_OV_DYNAMIC_DIR)
 */
struct proc_dir_entry *proc_symlink(const char *name,
		struct proc_dir_entry *parent, const char *dest)
{
	struct proc_dir_entry *ent;

	ent = proc_create(&parent,name,
			  (S_IFLNK | S_IRUGO | S_IWUGO | S_IXUGO),1,
			  sizeof(struct proc_dir_entry));

	if (ent) {
		ent->data = kmalloc((ent->size=strlen(dest))+1, GFP_KERNEL);
		if (ent->data) {
			strcpy((char*)ent->data,dest);
			proc_register(parent, ent);
		} else {
			kfree(ent);
			ent = NULL;
		}
	}
  	return ent;
  }

/*
 * Does not find dynamic entries (type == PROC_OV_DYNAMIC_DIR)
 */
struct proc_dir_entry *proc_mknod(const char *name, mode_t mode,
		struct proc_dir_entry *parent, kdev_t rdev)
{
	struct proc_dir_entry *ent;
  
	ent = proc_create(&parent,name,mode,1,sizeof(struct proc_dir_entry));
	if (ent) {
		ent->rdev = rdev;
		proc_register(parent, ent);
	}
	return ent;
}

/*
 * Does not find dynamic entries (type == PROC_OV_DYNAMIC_DIR)
 */
struct proc_dir_entry *proc_mkdir(const char *name, struct proc_dir_entry *parent)
{
	struct proc_dir_entry *ent;
  
	ent = proc_create(&parent,name,
			  (S_IFDIR | S_IRUGO | S_IXUGO),2, 
			  sizeof(struct proc_dir_entry));
	if (ent) {
		ent->proc_fops = &proc_dir_operations;
		ent->proc_iops = &proc_dir_inode_operations;
  
		proc_register(parent, ent);
	}
	return ent;
}

/*
 * For onevalue.c: create_proc_entry that allows larger structs
 */
struct proc_dir_entry *create_proc_entry_common(const char *name, 
						mode_t mode,
						int structlen, 
						struct proc_dir_entry *parent,
						unsigned int inos)
{
	struct proc_dir_entry *ent;
	nlink_t nlink;
  
  	if (S_ISDIR(mode)) {
  		if ((mode & S_IALLUGO) == 0)
			mode |= S_IRUGO | S_IXUGO;
		nlink = 2;
  	} else {
  		if ((mode & S_IFMT) == 0)
  			mode |= S_IFREG;
  		if ((mode & S_IALLUGO) == 0)
  			mode |= S_IRUGO;
		nlink = 1;
  	}
 
	ent = proc_create(&parent,name,mode,nlink,structlen);
	if (ent) {
		if (S_ISDIR(mode)) {
			ent->proc_fops = &proc_dir_operations;
			ent->proc_iops = &proc_dir_inode_operations;
		}
		proc_register(parent, ent);
	}
  	return ent;
}

struct proc_dir_entry *create_proc_entry(const char *name, mode_t mode,
					 struct proc_dir_entry *parent)
{
	return create_proc_entry_common(name, 
					mode, 
					sizeof(struct proc_dir_entry), 
					parent, 
					1);
}

void free_proc_entry(struct proc_dir_entry *de, unsigned int ino)
{
	if (ino < PROC_DYNAMIC_FIRST ||
		 ino >= PROC_DYNAMIC_FIRST+PROC_NDYNAMIC)
		return;
	if (S_ISLNK(de->mode) && de->data)
		kfree(de->data);
	kfree(de);
}

void proc_unregister_entry(struct proc_dir_entry *de)
{
	if (S_ISDIR(de->mode))
		de->parent->nlink--;
	free_inode_range(de->low_ino, get_inode_space(de));
	proc_kill_inodes(de);
	de->nlink = 0;
	if (!atomic_read(&de->count))
		free_proc_entry(de, de->low_ino);
	else {
		de->deleted = 1;
		printk("remove_proc_entry: %s/%s busy, count=%d\n",
		       de->parent->name, de->name, atomic_read(&de->count));
	}
}

/*
 * Remove a /proc entry and free it if it's not currently in use.
 * If it is in use, we set the 'deleted' flag.
 * Cannot be used to delete dynamic proc entries.
 */
void remove_proc_entry(const char *name, struct proc_dir_entry *parent)
{
	struct proc_dir_entry **p;
	struct proc_dir_entry *de;
	const char *fn = name;
	int len;

	if (!parent && xlate_proc_name(name, &parent, &fn) != 0)
		goto out;
	len = strlen(fn);
	for (p = &parent->subdir; *p; p=&(*p)->next ) {
		if (!proc_match(len, fn, *p))
			continue;
		de = *p;
		*p = de->next;
		de->next = NULL;
		proc_unregister_entry(de);
		break;
	}
out:
	return;
}

void __init proc_generic_init(void) 
{
	struct inode_range *r = kmalloc(sizeof(*r), GFP_KERNEL);
	if (!r) {
		printk(KERN_ERR "proc: failure allocating memory"
		       " at initialization!");
		return;
	}
	r->first = PROC_DYNAMIC_FIRST;
	r->end = r->first + PROC_NDYNAMIC;
	list_add(&r->free_inodes, &free_inodes);
}
