/*
 * fs/proc/onevalue.c --- one-value-per-file proc api
 *
 * Contains API implementations for "one-value-per-file" in the /proc
 * filesystem. Makes it possible to create directories dynamically
 * (like the per-process dirs /proc/<number>/). 
 *
 * When using this API, do not mix static entries that have been created 
 * with the old API with those created by the proc_ov functions, otherwise
 * the allocation of inode numbers may not work. It is ok to put ov files or
 * directories into static directories, but you do not insert static entries 
 * into ov directories.
 * As all kmalloc invocations use GFP_KERNEL you must not use the proc_ov
 * functions at interrupt time (hard or soft).
 */

#include <linux/proc_fs.h>
#include <linux/string.h>

/**
 * proc_ov_get_entry - Returns the entry for the given path.
 * @path: The path of the entry (may begin with "/proc", a relative path
 *        is interpreted as relative to "/proc/").
 * 
 * Returns a pointer to the corresponding struct proc_dir_entry, or %NULL if 
 * not found.
 */
struct proc_dir_entry *
proc_ov_get_entry(const char *path)
{
	const char *n;
	const char *name = path;
	struct proc_dir_entry *e = &proc_root;
	
	if (strstr(name, "/proc") == name)
		name += strlen("/proc");
	if (name[0] == '/')
		name++;
	if (!strlen(name))
		return e;

	while(1) {
		n = strchr(name, '/');
		if (!n)
			n = &name[strlen(name)];

		for (e = e->subdir; e; e = e->next) 
			if (proc_match(n - name, name, e))
				break;
		if (!e)
			return NULL;
		if (!(n[0] && n[1]))
			break;
		name = &n[1];
	}
	return e;
}

static int _get_ino_space(struct proc_dir_entry *p) 
{
	if (p->type == PROC_OV_DIR)
		return ((struct proc_ov_dir*)p)->ino_space;
	if (p->type == PROC_OV_DYNAMIC_DIR)
		return ((struct proc_ov_dynamic_dir*)p)->ino_space;
      	return 1;
}

/**
 * proc_ov_register_string - Adds a file containing a string.
 * @parent: The parent directory of the new file, use %NULL for root.
 * @name: The name of the new file.
 * @cb: The callback function to generate the file's string.
 * 
 * The parent must have been created using proc_ov_register_dir() or 
 * proc_ov_register_dynamic_dir().
 * Returns a pointer to the created struct proc_dir_entry, or %NULL if 
 * the creation failed or CONFIG_PROC_FS in not enabled.
 */
struct proc_dir_entry *
proc_ov_register_string(struct proc_dir_entry *parent, 
			const char *name,
			proc_ov_get_string_t cb) 
{
	struct proc_ov_value_string * ov;

	if (!parent)
		parent = &proc_root;

	ov = (struct proc_ov_value_string*) 
		create_proc_entry_common(name, 
					 0, 
					 sizeof(struct proc_ov_value_string),
					 parent,
					 _get_ino_space(parent));
	if (!ov)
		return NULL;

	ov->super.type = PROC_OV_STRING;
	ov->get_string = cb;
	return &ov->super;
}

/**
 * proc_ov_register_int - Adds a file containing an integer.
 * @parent: The parent directory of the new file. Use %NULL for root.
 * @name: The name of the new file.
 * @cb: The callback function to generate the file's integer.
 * 
 * The parent must have been created using proc_ov_register_dir() or 
 * proc_ov_register_dynamic_dir().
 * Returns a pointer to the created struct proc_dir_entry, or %NULL if 
 * the creation failed or CONFIG_PROC_FS in not enabled.
 */
struct proc_dir_entry *
proc_ov_register_int(struct proc_dir_entry *parent, 
		     const char *name,
		     proc_ov_get_int_t cb)
{
	struct proc_ov_value_int * ov;

	if (!parent)
		parent = &proc_root;

	ov = (struct proc_ov_value_int*) 
		create_proc_entry_common(name, 
					 0, 
					 sizeof(struct proc_ov_value_int),
					 parent,
					 _get_ino_space(parent));
	if (!ov)
		return NULL;

	ov->super.type = PROC_OV_INT;
	ov->get_int = cb;
	return &ov->super;	
}

/**
 * proc_ov_register_ulong - Adds a file containing an unsingned long.
 * @parent: The parent directory of the new file. Use %NULL for root.
 * @name: The name of the new file.
 * @cb: The callback function to generate the file's integer.
 * 
 * The parent must have been created using proc_ov_register_dir() or 
 * proc_ov_register_dynamic_dir().
 * Returns a pointer to the created struct proc_dir_entry, or %NULL if 
 * the creation failed or CONFIG_PROC_FS in not enabled.
 */
struct proc_dir_entry *
proc_ov_register_ulong(struct proc_dir_entry *parent, 
		       const char *name,
		       proc_ov_get_ulong_t cb)
{
	struct proc_ov_value_ulong * ov;

	if (!parent)
		parent = &proc_root;

	ov = (struct proc_ov_value_ulong*) 
		create_proc_entry_common(name, 
					 0, 
					 sizeof(struct proc_ov_value_ulong),
					 parent,
					 _get_ino_space(parent));
	if (!ov)
		return NULL;

	ov->super.type = PROC_OV_ULONG;
	ov->get_ulong = cb;
	return &ov->super;	
}


/**
 * proc_ov_register_enum - Adds a file containing an enum.
 * @parent: The parent directory of the new file. Use %NULL for the root 
 * directory.
 * @name: The name of the new file.
 * @enum_values: A string array containing the possible values. Must be valid
 *               as long as the file is registered and NULL-terminated.
 * @cb: The callback function to generate the file's value.
 * 
 * The parent must have been created using proc_ov_register_dir() or 
 * proc_ov_register_dynamic_dir().
 * Returns a pointer to the created struct proc_dir_entry, or %NULL if 
 * the creation failed or CONFIG_PROC_FS in not enabled.
 */
struct proc_dir_entry *
proc_ov_register_enum(struct proc_dir_entry *parent, 
		      const char *name,
		      const char **enum_values,
		      proc_ov_get_enum_t cb)
{
	struct proc_ov_value_enum * ov;

	if (!parent)
		parent = &proc_root;

	ov = (struct proc_ov_value_enum*) 
		create_proc_entry_common(name, 
					 0, 
					 sizeof(struct proc_ov_value_enum),
					 parent,
					 _get_ino_space(parent));
	if (!ov) 
		return NULL;

	ov->super.type = PROC_OV_ENUM;
	ov->get_enum = cb;
	ov->enum_values = enum_values;
	return &ov->super;
}

/**
 * proc_ov_register_dir - Adds a directory to the proc filesystem (with context support). 
 * @parent: The parent directory of the new dir. Use %NULL for the root directory.
 * @name: The name of the directory.
 * @denter: The callback function to generate the directory's context, and/or do
 *          things like locking. Can be %NULL.
 * @dleave: The callback function to do context-related cleanup work 
 *          (e.g. unlocking). Can be %NULL.
 * 
 * Returns a pointer to the created struct proc_dir_entry, or %NULL if 
 * the creation failed or CONFIG_PROC_FS in not enabled.
 */
struct proc_dir_entry * 
proc_ov_register_dir(struct proc_dir_entry *parent, 
		     const char *name,
		     proc_ov_dir_enter_t denter,
		     proc_ov_dir_leave_t dleave)
{
	struct proc_ov_dir * ov;
	int inos;

	if (!parent)
		parent = &proc_root;

	inos = _get_ino_space(parent);

	ov = (struct proc_ov_dir*) 
		create_proc_entry_common(name, 
					 S_IFDIR | S_IRUGO | S_IXUGO, 
					 sizeof(struct proc_ov_dir),
					 parent,
					 inos);
	if (!ov)
		return NULL;

	ov->super.type = PROC_OV_DIR;
	ov->enter = denter;
	ov->leave = dleave;
	ov->ino_space = inos;
	return &ov->super;
}


/**
 * proc_ov_register_dynamic_dir - Adds dynamic directories to a proc dir.
 * @parent: The parent directory of the dynamic directories. It is not 
 *          possible to add a dynamic dir to the root directory.
 * @size: The maximum number of entries in this directory. 
 * @dnum: A function that returns the number of directories.
 * @dname: A function that writes the name of the entry into the NAME_MAX long
 *         buffer. Can be %NULL, then the index number is the name.
 * @denter: The function to generate the directory's context, and/or 
 *          do things like locking. Can be %NULL.
 * @dleave: The function to do context-related cleanup work (e.g. unlocking). 
 *          Can be %NULL.
 * 
 * Returns a pointer to the created struct proc_dir_entry, or %NULL if 
 * the creation failed or CONFIG_PROC_FS in not enabled.
 */
struct proc_dir_entry *
proc_ov_register_dynamic_dir(struct proc_dir_entry *parent,
			     unsigned int size, 
			     proc_ov_dir_num_t dnum,
			     proc_ov_dir_name_t dname,
			     proc_ov_dir_context_t denter,
			     proc_ov_dir_leave_t dleave)
{
	int ino_space;
	struct proc_dir_entry *ent;
	struct proc_ov_dynamic_dir *ov;

	ov = (struct proc_ov_dynamic_dir*) 
		kmalloc(sizeof(struct proc_ov_dynamic_dir), GFP_KERNEL);
	if (!ov)
		return NULL;
	memset(ov, 0, sizeof(*ov));

	ino_space = _get_ino_space(parent) * size;

	ent = &ov->super;
	ent->type = PROC_OV_DYNAMIC_DIR;
	ent->name = "(dynamic dir stub)";
	ent->namelen = 0;
	ent->nlink = 2;
	ent->mode = S_IFDIR | S_IRUGO | S_IXUGO;
	if (!proc_init_dyn_dir(ent, ino_space)) {
		kfree(ov);
		return NULL;
	}

	ov->get_num = dnum;
	ov->get_name = dname;
	ov->get_context = denter;
	ov->leave = dleave;
	ov->max_size = size;
	ov->ino_space = ino_space;

	ent->parent = parent;
	ent->next = &parent->dyndir->super;
	parent->dyndir = ov;
	return ent;
}

static void remove_level(struct proc_dir_entry *first)
{
	struct proc_dir_entry *e = first;

	while (e) {
		proc_ov_remove(e);
		e = e->next;
	}
}

/**
 * proc_ov_register_remove - Removes the given entry, including its
 * children.
 *
 * @e: The entry to remove. Can be %NULL, will be ignored then.
 */
void proc_ov_remove(struct proc_dir_entry *e)
{
	struct proc_dir_entry **i = NULL;

	if (!e)
		return;

	remove_level(e->subdir);
	remove_level(&e->dyndir->super);

	if (e->type == PROC_OV_DYNAMIC_DIR) 
		i = (struct proc_dir_entry**) &e->parent->dyndir;
	else
		i = &e->parent->subdir;

	while (i && *i) {
		if (*i == e) {
			*i = (*i)->next;
			break;
		}
		i = &((*i)->next);
	}
	
	proc_unregister_entry(e);
}

