#ifndef _LINUX_PROC_FS_H
#define _LINUX_PROC_FS_H

#include <linux/config.h>
#include <linux/slab.h>

/*
 * The proc filesystem constants/structures
 */

/*
 * Offset of the first process in the /proc root directory..
 */
#define FIRST_PROCESS_ENTRY 256


/*
 * We always define these enumerators
 */

enum {
	PROC_ROOT_INO = 1,
};

/* Finally, the dynamically allocatable proc entries are reserved: */

#define PROC_DYNAMIC_FIRST 0x10000000
#define PROC_NDYNAMIC      0x0fffffff

#define PROC_SUPER_MAGIC 0x9fa0

/* buffer size is one page but our output routines use some slack for overruns */
#define PROC_BLOCK_SIZE	(PAGE_SIZE - 1024)

/* Maximum number of stacked contexts */
#define PROC_OV_MAX_CONTEXT_DEPTH 16

/*
 * This is not completely implemented yet. The idea is to
 * create an in-memory tree (like the actual /proc filesystem
 * tree) of these proc_dir_entries, so that we can dynamically
 * add new files to /proc.
 *
 * The "next" pointer creates a linked list of one /proc directory,
 * while parent/subdir create the directory structure (every
 * /proc file has a parent, but "subdir" is NULL for all
 * non-directory entries).
 *
 * "get_info" is called at "read", while "owner" is used to protect module
 * from unloading while proc_dir_entry is in use
 */

typedef	int (read_proc_t)(char *page, char **start, off_t off,
			  int count, int *eof, void *data);
typedef	int (write_proc_t)(struct file *file, const char *buffer,
			   unsigned long count, void *data);
typedef int (get_info_t)(char *, char **, off_t, int);


/*
 * These callbacks can be registered using proc_ov_register_(string|int|enum)
 * and write the content of a single file in the /proc hierarchy. As the files
 * can be contained in dynamic directories and to avoid duplicate code each
 * functions gets a context that can help it to generate the content. The
 * context is set by the callbacks of the proc_ov_register_dir() and
 * proc_ov_register_dynamic_dir(), in a normal static directory that has been 
 * created by create_proc_entry() function the context is NULL. The callbacks
 * must be prepared that the context can be NULL (because of race conditions). 
 * If it is and they dont expect the NULL they should return an error code.
 * All functions must return a positive value for success or a negative error 
 * code. They will be called in user context.
 * 
 * - proc_ov_get_string_t writes a string into buf. The size of the buffer is 
 *   PROC_BLOC_SIZE. The return value is the length of the string.
 * - proc_ov_get_int_t writes a single integer value to the buffer.
 * - proc_ov_get_ulong_t writes a unsigned long value to the buffer.
 * - proc_ov_get_enum_t writes the integer index of one of the string that have 
 *   been specified by proc_ov_register_enum().
 */
typedef int (proc_ov_get_string_t)(char *buf, 
				   void *context,
				   void *data);
typedef int (proc_ov_get_int_t)(int *value, 
				void *context,
				void *data);
typedef int (proc_ov_get_ulong_t)(unsigned long *value, 
				  void *context,
				  void *data);
typedef int (proc_ov_get_enum_t)(int *value, 
				 void *context,
				 void *data);

/*
 * These callbacks create the context for the proc_ov_get_* functions and
 * the names of the subdirectories of a dynamic directory.
 * They will be called in user context.
 * 
 * - proc_ov_dir_enter_t returns the context of a directory. The argument 
 *   old_config is the context of the parent director (NULL if there is no
 *   parent directory) The function can also be used for acquire locks.
 *
 * - proc_ov_dir_leave_t is called when a directory is left. The argument
 *   old_config contains the context of the parent context, new_context
 *   contains the context created by the matching proc_ov_dir_enter_t or 
 *   proc_ov_dir_context_t. 
 *
 * - proc_ov_dir_num_t is called in dynamic directories and returns the 
 *   number of subdirectories.
 *
 * - proc_ov_dir_name_t writes the name of the directory with the given 
 *   index into the given buffer (size: NAME_MAX) and returns the length
 *   of the string, or a negative value for failure. Must be prepared
 *   that index may be out-of-range, and then return an error.
 *
 * - proc_ov_dir_context_t returns the context of the directory with the 
 *   given index or NULL for failure.Must be prepared that index may be 
 *   out-of-range, and then return NULL.
 */
typedef void *(proc_ov_dir_enter_t)(void *old_context,
				    void *data);
typedef void (proc_ov_dir_leave_t)(void *old_context,
				   void *new_context,
				   void *data);

typedef int (proc_ov_dir_num_t)(void *context,
				void *data);
typedef int (proc_ov_dir_name_t)(int index,
				 void *context,
				 void *data,
				 char *name_buf);
typedef void *(proc_ov_dir_context_t)(int index,
				      void *old_context,
				      void *data);

struct proc_dir_entry {
	unsigned int low_ino;
	unsigned short namelen;
	const char *name;
	mode_t mode;
	nlink_t nlink;
	uid_t uid;
	gid_t gid;
	unsigned long size;
	struct inode_operations * proc_iops;
	struct file_operations * proc_fops;
	get_info_t *get_info;
	struct module *owner;
	struct proc_dir_entry *next, *parent, *subdir;
	struct proc_ov_dynamic_dir *dyndir;
	void *data;
	read_proc_t *read_proc;
	write_proc_t *write_proc;
	atomic_t count;		/* use count */
	int deleted;		/* delete flag */
	kdev_t	rdev;
	enum {
		PROC_STATIC_ENTRY = 0,
		PROC_OV_DIR,
		PROC_OV_DYNAMIC_DIR,
		PROC_OV_STRING,
		PROC_OV_INT,
		PROC_OV_ULONG,
		PROC_OV_ENUM
	} type;
};

struct proc_ov_dir {
	struct proc_dir_entry super;
	proc_ov_dir_enter_t *enter;
	proc_ov_dir_leave_t *leave;
	int ino_space;
};

struct proc_ov_dynamic_dir {
	struct proc_dir_entry super;
	proc_ov_dir_context_t *get_context;
	proc_ov_dir_leave_t *leave; 
	proc_ov_dir_num_t *get_num;
	proc_ov_dir_name_t *get_name;
	int max_size;
	int ino_space;
};

struct proc_ov_value_string {
	struct proc_dir_entry super;
	proc_ov_get_string_t *get_string;
};

struct proc_ov_value_int {
	struct proc_dir_entry super;
	proc_ov_get_int_t *get_int;
};

struct proc_ov_value_ulong {
	struct proc_dir_entry super;
	proc_ov_get_ulong_t *get_ulong;
};

struct proc_ov_value_enum {
	struct proc_dir_entry super;
	proc_ov_get_enum_t *get_enum;
	const char **enum_values;
};

#define PROC_INODE_PROPER(inode) (((inode)->i_ino > 0xfff) && \
                                  ((inode)->i_ino < PROC_DYNAMIC_FIRST))

#ifdef CONFIG_PROC_FS

extern struct proc_dir_entry proc_root;
extern struct proc_dir_entry *proc_root_fs;
extern struct proc_dir_entry *proc_net;
extern struct proc_dir_entry *proc_bus;
extern struct proc_dir_entry *proc_root_driver;
extern struct proc_dir_entry *proc_root_kcore;

extern void proc_root_init(void);
extern void proc_generic_init(void);
extern void proc_misc_init(void);

struct dentry *proc_pid_lookup(struct inode *dir, struct dentry * dentry);
void proc_pid_delete_inode(struct inode *inode);
int proc_pid_readdir(struct file * filp, void * dirent, filldir_t filldir);

extern struct proc_dir_entry *create_proc_entry(const char *name, mode_t mode,
						struct proc_dir_entry *parent);
extern void remove_proc_entry(const char *name, struct proc_dir_entry *parent);

extern struct vfsmount *proc_mnt;
extern struct super_block *proc_read_super(struct super_block *,void *,int);
extern struct inode * proc_get_inode(struct super_block *, unsigned int, struct proc_dir_entry *);

extern int proc_match(int, const char *,struct proc_dir_entry *);

/*
 * These are generic /proc routines that use the internal
 * "struct proc_dir_entry" tree to traverse the filesystem.
 *
 * The /proc root directory has extended versions to take care
 * of the /proc/<pid> subdirectories.
 */
extern int proc_readdir(struct file *, void *, filldir_t);
extern struct dentry *proc_lookup(struct inode *, struct dentry *);

extern struct file_operations proc_kcore_operations;
extern struct file_operations proc_kmsg_operations;
extern struct file_operations ppc_htab_operations;

/*
 * onevalue.c
 */
extern struct proc_dir_entry *
proc_ov_get_entry(const char *path);

extern struct proc_dir_entry *
proc_ov_register_string(struct proc_dir_entry *parent, 
			const char *name,
			proc_ov_get_string_t *cb);

extern struct proc_dir_entry *
proc_ov_register_int(struct proc_dir_entry *parent, 
		     const char *name,
		     proc_ov_get_int_t *cb);

extern struct proc_dir_entry *
proc_ov_register_ulong(struct proc_dir_entry *parent, 
		       const char *name,
		       proc_ov_get_ulong_t *cb);

extern struct proc_dir_entry *
proc_ov_register_enum(struct proc_dir_entry *parent, 
		      const char *name,
		      const char **enum_values,
		      proc_ov_get_enum_t *cb);

extern struct proc_dir_entry * 
proc_ov_register_dir(struct proc_dir_entry *parent, 
		     const char *name,
		     proc_ov_dir_enter_t *denter,
		     proc_ov_dir_leave_t *dleave);

extern struct proc_dir_entry *
proc_ov_register_dynamic_dir(struct proc_dir_entry *parent,
			     unsigned int size, 
			     proc_ov_dir_num_t *dnum,
			     proc_ov_dir_name_t *dname,
			     proc_ov_dir_context_t *denter,
			     proc_ov_dir_leave_t *dleave);

extern void proc_ov_remove(struct proc_dir_entry *e);

/*
 * helper from generic.c for onevalue.c
 */
extern int proc_init_dyn_dir(struct proc_dir_entry *, unsigned int);
extern struct proc_dir_entry *
create_proc_entry_common(const char *, mode_t, int, struct proc_dir_entry *, unsigned int);
extern void proc_unregister_entry(struct proc_dir_entry *);

/*
 * proc_tty.c
 */
struct tty_driver;
extern void proc_tty_init(void);
extern void proc_tty_register_driver(struct tty_driver *driver);
extern void proc_tty_unregister_driver(struct tty_driver *driver);

/*
 * proc_devtree.c
 */
extern void proc_device_tree_init(void);

/*
 * proc_rtas.c
 */
extern void proc_rtas_init(void);

extern struct proc_dir_entry *proc_symlink(const char *,
		struct proc_dir_entry *, const char *);
extern struct proc_dir_entry *proc_mknod(const char *,mode_t,
		struct proc_dir_entry *,kdev_t);
extern struct proc_dir_entry *proc_mkdir(const char *,struct proc_dir_entry *);

static inline struct proc_dir_entry *create_proc_read_entry(const char *name,
	mode_t mode, struct proc_dir_entry *base, 
	read_proc_t *read_proc, void * data)
{
	struct proc_dir_entry *res=create_proc_entry(name,mode,base);
	if (res) {
		res->read_proc=read_proc;
		res->data=data;
	}
	return res;
}
 
static inline struct proc_dir_entry *create_proc_info_entry(const char *name,
	mode_t mode, struct proc_dir_entry *base, get_info_t *get_info)
{
	struct proc_dir_entry *res=create_proc_entry(name,mode,base);
	if (res) res->get_info=get_info;
	return res;
}
 
static inline struct proc_dir_entry *proc_net_create(const char *name,
	mode_t mode, get_info_t *get_info)
{
	return create_proc_info_entry(name,mode,proc_net,get_info);
}

static inline void proc_net_remove(const char *name)
{
	remove_proc_entry(name,proc_net);
}

#else

#define proc_root_driver NULL

static inline struct proc_dir_entry *proc_net_create(const char *name, mode_t mode, 
	get_info_t *get_info) {return NULL;}
static inline void proc_net_remove(const char *name) {}

static inline struct proc_dir_entry *create_proc_entry(const char *name,
	mode_t mode, struct proc_dir_entry *parent) { return NULL; }

static inline void remove_proc_entry(const char *name, struct proc_dir_entry *parent) {};
static inline struct proc_dir_entry *proc_symlink(const char *name,
		struct proc_dir_entry *parent,char *dest) {return NULL;}
static inline struct proc_dir_entry *proc_mknod(const char *name,mode_t mode,
		struct proc_dir_entry *parent,kdev_t rdev) {return NULL;}
static inline struct proc_dir_entry *proc_mkdir(const char *name,
	struct proc_dir_entry *parent) {return NULL;}

static inline struct proc_dir_entry *create_proc_read_entry(const char *name,
	mode_t mode, struct proc_dir_entry *base, 
	int (*read_proc)(char *, char **, off_t, int, int *, void *),
	void * data) { return NULL; }
static inline struct proc_dir_entry *create_proc_info_entry(const char *name,
	mode_t mode, struct proc_dir_entry *base, get_info_t *get_info)
	{ return NULL; }

static inline void proc_tty_register_driver(struct tty_driver *driver) {};
static inline void proc_tty_unregister_driver(struct tty_driver *driver) {};

extern struct proc_dir_entry proc_root;

static inline struct proc_dir_entry *proc_ov_get_entry(const char *path) 
{ return NULL; }
static inline struct proc_dir_entry *proc_ov_register_string(
	struct proc_dir_entry *parent, const char *name, 
	proc_ov_get_string_t *cb) { return NULL; }
static inline struct proc_dir_entry *proc_ov_register_int(
	struct proc_dir_entry *parent, const char *name, 
	proc_ov_get_int_t *cb) { return NULL; }
static inline struct proc_dir_entry *proc_ov_register_ulong(
	struct proc_dir_entry *parent, const char *name, 
	proc_ov_get_ulong_t *cb) { return NULL; }
static inline struct proc_dir_entry *proc_ov_register_enum(
	struct proc_dir_entry *parent, const char *name, 
	const char **enum_values, proc_ov_get_enum_t *cb) { return NULL; }
static inline struct proc_dir_entry *proc_ov_register_dir(
	struct proc_dir_entry *parent, const char *name, 
	proc_ov_dir_enter_t *denter, proc_ov_dir_leave_t *dleave)
{ return NULL; }
static inline struct proc_dir_entry *proc_ov_register_dynamic_dir(
	struct proc_dir_entry *parent, unsigned int size, 
	proc_ov_dir_num_t *dnum, proc_ov_dir_name_t *dmatch, 
	proc_ov_dir_context_t *denter, proc_ov_dir_leave_t *dleave) 
{ return NULL; }
static inline void proc_ov_remove(struct proc_dir_entry *e) {}

#endif /* CONFIG_PROC_FS */

#endif /* _LINUX_PROC_FS_H */
