#ifndef _LINUX_VECTOR_H
#define _LINUX_VECTOR_H

/*
 * Implementation of a dynamic-sized vector, can also be used as a set of 
 * unique element where the order doesn't matter. Should be used for small
 * data types like integers and pointers, use list.h for complex structures.
 * Uses kmalloc/kfree for memory allocation. 
 */

#include <linux/slab.h>
#include <linux/string.h>

/* mimimum increase for the vector */
#define VECTOR_MIN_INCREASE 8


struct vector {
	unsigned int capacity;
	unsigned int size;
	void *data;
};

static inline int __init_vector(struct vector *v, 
				unsigned int typesize,
				unsigned int capacity,
				int kmallocflags)
{
        if (!capacity)
		v->capacity = VECTOR_MIN_INCREASE;
	else
		v->capacity = capacity; 
	v->size = 0;
	v->data = kmalloc(v->capacity*typesize, kmallocflags);
	return v->data ? -1 : 0;
}

/*
 * Initial the vector or type with the given initial capacity.
 * Returns nonzero for success or 0 for failure (kmalloc failed).
 * @vector: a ptr to the struct vector
 * @type: the type of the vector elements
 * @capacity: the initial capacity (use 0 if you don't know)
 * @kmallocflags: flags for kmalloc (use GFP_KERNEL for normal
 *                 memory, GFP_ATOMIC for interrupt handlers)
 */
#define INIT_VECTOR(vector, type, capacity, kmallocflags) \
        __init_vector(vector, sizeof(type), capacity, kmallocflags)


static inline void __vector_free(struct vector *v) 
{
	kfree(v->data);
}

/*
 * Frees the resources allocated by the given vector.
 * @vector: a ptr to the struct vector
 */	
#define FREE_VECTOR(vector) __vector_free(vector)

static int __vector_increase_capacity(struct vector *v,
				      unsigned int element_size,
				      int kmallocflags)
{
	void *newdata;
	unsigned int newcapacity ;

	newcapacity = v->capacity + VECTOR_MIN_INCREASE;
	newdata = kmalloc(newcapacity * element_size, kmallocflags);
	if (!newdata)
		return 0;
	memcpy(newdata, v->data, v->size * element_size);
	kfree(v->data);
	v->data = newdata;
	v->capacity = newcapacity;
	return 1;
}

static inline int __vector_add(struct vector *v,
			       void *element,
			       unsigned int element_size,
			       int kmallocflags)
{
	if (v->size >= v->capacity)
		if (!__vector_increase_capacity(v, element_size, kmallocflags))
			return 0;

	memcpy(v->data + (v->size*element_size), 
	       element, 
	       element_size);
	v->size++;
	return -1;
}


/*
 * Adds the given element to the vector. 
 * Returns nonzero for success or 0 for failure (kmalloc failed).
 * @vector: a ptr to the struct vector
 * @elementptr: the pointer to the element to add
 * @type: the type of the vector elements
 * @kmallocflags: flags for kmalloc (use GFP_KERNEL for normal
 *                 memory, GFP_ATOMIC for interrupt handlers)
 */
#define VECTOR_ADD(vector, elementptr, type, kmallocflags) \
        __vector_add(vector, elementptr, sizeof(type), kmallocflags)



static void __vector_shrink(struct vector *v,
			    void *part1,
			    unsigned int part1size,
			    void *part2,
			    unsigned int part2size,
			    unsigned int element_size,
			    int kmallocflags)
{
	void *newdata;
	unsigned int newcapacity;

	newcapacity = part1size + part2size + VECTOR_MIN_INCREASE;
	newdata = kmalloc(newcapacity * element_size, kmallocflags);
	if (!newdata) {
		/* if kmalloc fails, just dont change the size, no error */
		if (part1 != v->data) 
			memmove(v->data, part1, part1size * element_size);
		memmove(v->data + (part1size * element_size),  
			part2, 
			part2size * element_size);
		v->size = part1size + part2size;
		return;
	}
	memcpy(newdata, part1, part1size * element_size);
	memcpy(newdata+(part1size * element_size), 
	       part2, 
	       part2size * element_size);
	kfree(v->data);
	v->data = newdata;	
	v->size = part1size + part2size;
	v->capacity = newcapacity;
}

static inline void __vector_remove_index(struct vector *v,
					 int index,
					 unsigned int element_size,
					 int kmallocflags)
{
	if ((v->size-1) < (v->capacity-(2*VECTOR_MIN_INCREASE))) {
		__vector_shrink(v, 
				v->data, 
				index, 
				v->data + ((index+1) * element_size),
				v->size-(index+1),
				element_size,
				kmallocflags);
		return;
	}

	memmove(v->data + (index * element_size),
		v->data + ((index+1) * element_size),
		(v->size-(index+1)) * element_size);
	v->size--;
}

static inline void __vector_remove_index_reorder(struct vector *v,
						 int index,
						 unsigned int element_size,
						 int kmallocflags)
{
	if ((v->size-1) < (v->capacity-(2*VECTOR_MIN_INCREASE))) {
		__vector_shrink(v, 
				v->data, 
				index, 
				v->data + ((index+1) * element_size),
				v->size-(index+1),
				element_size,
				kmallocflags);
		return;
	}

	memmove(v->data + (index * element_size),
		v->data + ((v->size-1) * element_size),
		element_size);
	v->size--;
}

/*
 * Removes the element at the given position from the vector,
 * keeps the order of the remaining elements.
 * @vector: a ptr to the struct vector
 * @index: the index of the element to remove
 * @type: the type of the vector elements
 * @kmallocflags: flags for kmalloc (use GFP_KERNEL for normal
 *                 memory, GFP_ATOMIC for interrupt handlers)
 */
#define VECTOR_REMOVE_INDEX(vector, index, type, kmallocflags) \
        __vector_remove_index(vector, index, sizeof(type), kmallocflags)

/*
 * Removes the element at the given position from the vector, this
 * may reorder the remaining elements. Faster than VECTOR_REMOVE_INDEX.
 * @vector: a ptr to the struct vector
 * @index: the index of the element to remove
 * @type: the type of the vector elements
 * @kmallocflags: flags for kmalloc (use GFP_KERNEL for normal
 *                 memory, GFP_ATOMIC for interrupt handlers)
 */
#define VECTOR_REMOVE_INDEX_REORDER(vector, index, type, kmallocflags) \
        __vector_remove_index_reorder(vector, index, sizeof(type), kmallocflags)


static inline int __vector_find_index(struct vector *v, 
				      void *element,
				      unsigned int element_size)
{
	int i;
	for (i = 0; i < v->size; i++)
		if (!memcmp(element, v->data+(i*element_size), element_size))
			return i;

	return -1;
}

static inline int __vector_remove_element(struct vector *v,
					  void *element,
					  unsigned int element_size,
					  int kmallocflags)
{
	int index = __vector_find_index(v, element, element_size);
	if (index < 0)
		return 1;
	__vector_remove_index(v, index, element_size, kmallocflags);
	return 0;
}

static inline int __vector_remove_element_reorder(struct vector *v,
						  void *element,
						  unsigned int element_size,
						  int kmallocflags)
{
	int index = __vector_find_index(v, element, element_size);
	if (index < 0)
		return 1;
	__vector_remove_index_reorder(v, index, element_size, kmallocflags);
	return 0;
}

/*
 * Removes the given element from the vector, will copy the following 
 * elements to the empty place.
 * Returns 0 for success or nonzero if the element has not been found
 * in the vector
 * @vector: a ptr to the struct vector
 * @element: a pointer to a copy of the element to remove
 * @type: the type of the vector elements
 * @kmallocflags: flags for kmalloc (use GFP_KERNEL for normal
 *                 memory, GFP_ATOMIC for interrupt handlers)
 */
#define VECTOR_REMOVE_ELEMENT(vector, element, type, kmallocflags) \
           __vector_remove_element(vector, element, sizeof(type), kmallocflags)

/*
 * Removes the given element from the vector. May reorder the following
 * elements, but is faster that VECTOR_REMOVE_ELEMENT.
 * Returns 0 for success or nonzero if the element has not been found
 * in the vector
 * @vector: a ptr to the struct vector
 * @element: a pointer to a copy of the element to remove
 * @type: the type of the vector elements
 * @kmallocflags: flags for kmalloc (use GFP_KERNEL for normal
 *                 memory, GFP_ATOMIC for interrupt handlers)
 */
#define VECTOR_REMOVE_ELEMENT_REORDER(vector, element, type, kmallocflags) \
           __vector_remove_element_reorder(vector, element, sizeof(type), kmallocflags)

/*
 * Returns the size of the vector.
 * @vector: a ptr to the struct vector
 */
#define VECTOR_GET_SIZE(vector) ((vector)->size)

/*
 * Returns a pointer to all elements of the array.
 * You can use the typeptr to get the element of the current element,
 * but you must not modify it.
 * @vector: a ptr to the struct vector
 * @type: the type of the vector elements
 */
#define VECTOR_GET_ARRAY(vector, type) \
        ((type*) (vector)->data)

/*
 * Returns a pointer to the element at the given index.
 * @vector: a ptr to the struct vector
 * @index: the index of the element
 * @type: the type of the vector elements
 */
#define VECTOR_GET_ELEMENTPTR_AT(vector, index, type) \
           (&VECTOR_GET_ARRAY(vector, type)[index])

/*
 * Returns the element at the given index.
 * @vector: a ptr to the struct vector
 * @index: the index of the element
 * @type: the type of the vector elements
 */
#define VECTOR_GET_ELEMENT_AT(vector, index, type) \
           (VECTOR_GET_ARRAY(vector, type)[index])

/*
 * Executes the following block once for each element in the vector.
 * You can use the typeptr to get the element of the current element,
 * but you must not modify it.
 * @vector: a ptr to the struct vector
 * @typeptr: some pointer of the element type that iterares over all elements
 * @type: the type of the vector elements
 */
#define VECTOR_FOR_EACH(vector, typeptr, type) \
        for (typeptr = (type*) (vector)->data; \
             ((void*)typeptr) < ((vector)->data + sizeof(type)*((vector)->size)); \
             typeptr++)


static inline int __vector_find_if(struct vector *v, 
				   unsigned int element_size, 
				   int (*compare)(void *elem, void *privdata),
				   void *privdata)
{
	int i;
	for (i = 0; i < v->size; i++)
		if(!compare(v->data+(i*element_size), privdata))
			return i;
	return -1;
}
/*
 * Returns the index of the first element that fulfills the given predicate 
 * functions or returns -1 if none found.
 * @vector: a ptr to the struct vector
 * @type: the type of the vector elements
 * @cmpfunc: a int (*compare)(void *element, void *privdata) function that
 *           compares the element to the private data and returns 0 of they
 *           are equal and non-zero otherwise
 * @privdata: the private data that will be given to cmpfunc
 * @type: the type of the vector elements
 */
#define VECTOR_FIND(vector, type, cmpfunc, privdata) \
        __vector_find_if(vector, sizeof(type), cmpfunc, privdata)


static inline int __vector_remove_func(struct vector *v, 
				       unsigned int element_size, 
				       int (*compare)(void *elem, void *privdata),
				       void *privdata,
				       int kmallocflags)
{
	int index = __vector_find_if(v, element_size, compare, privdata);
	if (index < 0)
		return -1;
	__vector_remove_index(v, index, element_size, kmallocflags);
	return index;
}
static inline int __vector_remove_func_reorder(struct vector *v, 
					       unsigned int element_size, 
					       int (*compare)(void *elem, void *privdata),
					       void *privdata,
					       int kmallocflags)
{
	int index = __vector_find_if(v, element_size, compare, privdata);
	if (index < 0)
		return -1;
	__vector_remove_index_reorder(v, index, element_size, kmallocflags);
	return index;
}
/*
 * Removes first element that is found using the remove func. Returns 
 * the index of the removed element or -1 if there was none.
 * @vector: a ptr to the struct vector
 * @type: the type of the vector elements
 * @cmpfunc: a int (*compare)(void *element, void *privdata) function that
 *           compares the element to the private data and returns 0 of they
 *           are equal and non-zero otherwise
 * @privdata: the private data that will be given to cmpfunc
 * @type: the type of the vector elements
 * @kmallocflags: flags for kmalloc (use GFP_KERNEL for normal
 *                 memory, GFP_ATOMIC for interrupt handlers)
 */
#define VECTOR_REMOVE_FUNC(vector, type, cmpfunc, privdata, kmallocflags) \
        __vector_remove_func(vector, sizeof(type), cmpfunc, privdata, kmallocflags)

/*
 * Removes first element that is found using the remove func, the elements
 * may be reordered. Returns the index of the removed element or -1
 * if there was none.
 * @vector: a ptr to the struct vector
 * @type: the type of the vector elements
 * @cmpfunc: a int (*compare)(void *element, void *privdata) function that
 *           compares the element to the private data and returns 0 of they
 *           are equal and non-zero otherwise
 * @privdata: the private data that will be given to cmpfunc
 * @type: the type of the vector elements
 * @kmallocflags: flags for kmalloc (use GFP_KERNEL for normal
 *                 memory, GFP_ATOMIC for interrupt handlers)
 */
#define VECTOR_REMOVE_FUNC_REORDER(vector, type, cmpfunc, privdata, kmallocflags) \
        __vector_remove_func_reorder(vector, sizeof(type), cmpfunc, privdata, kmallocflags)

#endif
